#  Copyright (c) Thomas Else 2023.
#  License: MIT

import unittest

import numpy as np

from patato import PreProcessor, PAData
from patato.core.image_structures.pa_time_data import PATimeSeries
from patato.data.get_example_datasets import get_msot_time_series_example

from patato.processing.gpu_preprocessing_algorithm import GPUMSOTPreProcessor
from patato.processing.preprocessing_algorithm import NumpyPreProcessor


class TestPreprocessing(unittest.TestCase):
    def setUp(self) -> None:
        self.pa = PAData.from_hdf5("test_data.hdf5")

    def _test_preprocessor(self, pre_processor):
        detectors_start = self.pa.get_scan_geometry()

        time_factor = 3
        detector_factor = 2

        start_time_series = self.pa.get_time_series()

        preproc = pre_processor(time_factor=time_factor, detector_factor=detector_factor)

        new_t, d, _ = preproc.run(start_time_series, self.pa)

        self.assertIsInstance(d, dict)
        self.assertIsInstance(new_t, PATimeSeries)
        self.assertEqual(new_t.shape, start_time_series.shape[:-2] + (start_time_series.shape[-2] * detector_factor,
                                                                      start_time_series.shape[-1] * time_factor))
        self.assertTrue(np.all(np.isclose(detectors_start[0], d["geometry"][0])))
        return new_t

    def test_gpu_processing(self):
        try:
            import cupy as cp
        except ImportError:
            return None  # Skip test if cupy is not installed
        new_t = self._test_preprocessor(GPUMSOTPreProcessor)
        print(np.mean(new_t[0, 0].values))

    def test_numpy_processing(self):
        new_t = self._test_preprocessor(NumpyPreProcessor)
        print(np.mean(new_t[0, 0].values))

    def test_jax_preprocessing(self):
        new_t = self._test_preprocessor(PreProcessor)
        print(np.mean(new_t[0, 0].values))



"""
Image sequence - abstract classes for processing datasets from PA data.
"""

#  Copyright (c) Thomas Else 2023.
#  License: MIT

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Tuple, Iterable

import numpy as np
import xarray

try:
    from pyopencl.array import Array
except ImportError:
    Array = None
try:
    import cupy as cp
except ImportError:
    cp = None

from xarray import DataArray

try:
    import jax.numpy as jnp
except ImportError:
    jnp = None

if jnp.DeviceArray not in xarray.core.variable.NON_NUMPY_SUPPORTED_ARRAY_TYPES:
    xarray.core.variable.NON_NUMPY_SUPPORTED_ARRAY_TYPES += (jnp.DeviceArray, )
from dask.array.core import Array as DaskArray

from ...io.attribute_tags import ReconAttributeTags
from ...processing.processing_algorithm import ProcessingResult
from ...utils.plotting import type_cmaps
from ...utils.rois.roi_type import ROI


def _get_matplotlib_scalebar_size(scalebar):
    # Mode 1: Auto
    ax = scalebar.axes
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    from matplotlib import rcParams

    def _get_value(attr, default):
        value = getattr(scalebar, attr)
        if value is None:
            value = rcParams.get("scalebar." + attr, default)
        return value

    rotation = _get_value("rotation", "horizontal").lower()
    length_fraction = _get_value("length_fraction", 0.2)
    fixed_value = scalebar.fixed_value
    fixed_units = scalebar.fixed_units or scalebar.units

    if rotation == "vertical":
        xlim, ylim = ylim, xlim

    if scalebar.fixed_value is None:
        length_px = abs(xlim[1] - xlim[0]) * length_fraction
        length_px, value, units = scalebar._calculate_best_length(length_px)
    else:
        value = fixed_value
        units = fixed_units

    return scalebar.scale_formatter(value, scalebar.dimension.to_latex(units))


class DataSequence(ProcessingResult, ABC):
    """
    Abstract base class for defining a sequence of data, e.g. raw data, reconstructed images, unmixed images. Enables
    consistent saving and processing for all of these data types.
    """
    n_im_dim = 3

    @property
    def attributes(self):
        return self.da.attrs

    def __init__(self, data, dimensions, coordinates=None, attributes=None, hdf5_sub_name=None,
                 algorithm_id=None):
        if coordinates is None:
            coordinates = {x: np.arange(data.shape[i]) for i, x in enumerate(dimensions)}
        ProcessingResult.__init__(self)
        self.da = DataArray(data=data, dims=dimensions, coords=coordinates,
                            attrs=attributes)
        self._cmap = None
        self.hdf5_sub_name = hdf5_sub_name
        self.algorithm_id = algorithm_id

    def __getitem__(self, item):
        c = self.copy()
        try:
            c.da = c.da[item]
        except TypeError:
            # Fudging the JAX implementation...
            c.da = c.da.copy()
            c.da.variable._data = np.asarray(c.da.variable._data)
            c.da = c.da[item]
        return c

    # TODO: Implement a concatenate function
    def copy(self, cls=None):
        from copy import copy
        c = copy(self)
        if cls is not None:
            c.__class__ = cls
        return c

    @property
    def shape_2d(self):
        if np.any(np.array(self.raw_data.shape[-3:]) == 1):
            return tuple([x for x in self.raw_data.shape[-3:] if x != 1])
        else:
            return self.raw_data.shape[-2:]

    @property
    def extent(self):
        coords = [self.da.coords[x] for x in self.two_dims()]
        return sum(((np.min(np.array(c)), np.max(np.array(c))) for c in coords), ())

    def two_dims(self):
        if self.da.coords["x"].size == 1:
            return "y", "z"
        elif self.da.coords["y"].size == 1:
            return "x", "z"
        else:
            return "x", "y"

    def to_2d(self):
        s = (0, ) * (len(self.shape) - self.n_im_dim)
        if self.n_im_dim > 2:
            slicer = [0] * self.n_im_dim
            for i in np.argsort(self.shape[-self.n_im_dim:])[-2:]:
                slicer[i] = slice(None)
            s += tuple(slicer)
        return self[s]

    def imshow(self, ax=None, roi_mask: Tuple["ROI", Iterable["ROI"]] = None,
               mask_roi=True,
               cmap=None, scale_kwargs=None, return_scalebar_dimension=False, scalebar=True, transpose=False, log=False,
               **kwargs):
        if scale_kwargs is None:
            scale_kwargs = {}

        import matplotlib.pyplot as plt
        if ax is None:
            ax = plt.gca()

        if roi_mask is not None:
            if type(roi_mask) is not ROI:
                try:
                    mask, image_slice = roi_mask[0].to_mask_slice(self)
                    image_slice = image_slice.to_2d()
                    display_image = np.squeeze(image_slice.numpy_array).astype(np.float64)
                    overall_mask = np.zeros(display_image.shape, dtype=bool)
                    for roi in roi_mask:
                        mask, _ = roi.to_mask_slice(self)
                        if mask_roi:
                            overall_mask[np.squeeze(mask)] = True
                    if mask_roi:
                        display_image[~overall_mask] = np.nan
                except TypeError:
                    raise ValueError("roi_mask must be a ROI or a tuple of ROIs")
            else:
                mask, image_slice = roi_mask.to_mask_slice(self)
                image_slice = image_slice.to_2d()
                display_image = np.squeeze(image_slice.numpy_array).astype(np.float64)
                if mask_roi:
                    display_image[~np.squeeze(mask)] = np.nan
        else:
            display_image = self.to_2d().numpy_array

        interpolation = "antialiased"
        if display_image.dtype == np.bool_:
            interpolation = "nearest"

        if cmap is None:
            cmap = self.cmap
        if np.iscomplexobj(display_image):
            display_image = np.real(display_image)
        display_image = np.squeeze(display_image)
        if "origin" not in kwargs:
            kwargs["origin"] = "lower"
        if transpose:
            extent = self.extent[2:] + self.extent[:2]
            display_image = display_image.T
        else:
            extent = self.extent
        
        if log:
            display_image = np.log(display_image)
        im = ax.imshow(display_image, extent=extent,
                       cmap=cmap, **kwargs, interpolation=interpolation)
        ax.axis("off")
        if scalebar:
            from matplotlib_scalebar.scalebar import ScaleBar
            scale_kwargs_defaults = dict(length_fraction=0.1, location="lower right",
                                         font_properties=dict(size="xx-small"), box_alpha=0., color="w")
            scale_kwargs_defaults.update(scale_kwargs)
            scalebar = ScaleBar(1, "m", **scale_kwargs_defaults)
            ax.add_artist(scalebar)
        if return_scalebar_dimension:
            return im, _get_matplotlib_scalebar_size(scalebar)
        else:
            return im

    @property
    def cmap(self):
        if self._cmap is None:
            return type_cmaps[self.get_hdf5_group_name()]
        else:
            return self._cmap

    @cmap.setter
    def cmap(self, x):
        self._cmap = x

    @staticmethod
    def get_ax1_label_meaning():
        return ""

    @abstractmethod
    def get_hdf5_group_name(self):
        pass

    @property
    def values(self):
        return self.raw_data

    @property
    def numpy_array(self):
        return np.asarray(self.values)

    @property
    def raw_data(self):
        # TODO: check if this has side effects in future. Previously, this said if type(self.da.data) == Array:
        if type(self.da.variable._data) == DaskArray:
            return np.array(self.da.data)
        else:
            return self.da.variable._data

    @raw_data.setter
    def raw_data(self, value):
        self.da.values = value

    @property
    def shape(self):
        return self.da.shape

    @property
    def ndim(self):
        return self.da.ndim

    @property
    def dtype(self):
        return self.da.dtype

    @property
    def ax_1_labels(self):
        r = np.asarray(self.da.coords.get(self.get_ax1_label_meaning(), []))
        if r.ndim == 0 and r.item() is None:
            return r.item()
        else:
            return r

    @property
    def ax_0_labels(self):
        return self.da.coords["frames"]

    @staticmethod
    def ax_0_exists():
        return True


class ImageSequence(DataSequence):
    @staticmethod
    def is_single_instance():
        return False

    def get_hdf5_group_name(self):
        raise NotImplementedError()

    def __add__(self, other):
        # A really lazy implementation of concatenating these datasets. There is 100% a better way to do this..
        new_data = xarray.concat([self.da, other.da], dim=other.da.dims[0])
        output = ImageSequence(new_data.values, self.ax_1_labels, self.algorithm_id, self.fov_3d, self.attributes,
                               self.hdf5_sub_name, ax1_meaning=self.get_ax1_label_meaning())
        output.__class__ = self.__class__
        return output

    @property
    def fov(self):
        n_pixel_tags = [ReconAttributeTags.X_NUMBER_OF_PIXELS,
                        ReconAttributeTags.Y_NUMBER_OF_PIXELS,
                        ReconAttributeTags.Z_NUMBER_OF_PIXELS]
        fov_tags = [ReconAttributeTags.X_FIELD_OF_VIEW,
                    ReconAttributeTags.Y_FIELD_OF_VIEW,
                    ReconAttributeTags.Z_FIELD_OF_VIEW]
        n_pixels = np.array([self.attributes.get(tag, 1) for tag in n_pixel_tags])
        if np.all(n_pixels == 1):
            # Old-style data
            n_pixels = np.array([self.attributes.get(ReconAttributeTags.OLD_RECON_NX)] * 2)
            fov_tags = [ReconAttributeTags.OLD_FIELD_OF_VIEW] * 2
        axes = np.where(~(n_pixels == 1))[0]
        fov_x = self.attributes.get(fov_tags[axes[0]], None)
        fov_y = self.attributes.get(fov_tags[axes[1]], None)
        return fov_x, fov_y

    @property
    def fov_3d(self):
        fov_tags = [ReconAttributeTags.X_FIELD_OF_VIEW,
                    ReconAttributeTags.Y_FIELD_OF_VIEW,
                    ReconAttributeTags.Z_FIELD_OF_VIEW]
        fov = np.array([self.attributes.get(tag, None) for tag in fov_tags])
        if all([x is None for x in fov_tags]):
            # Old-style data
            n_pixels = self.raw_data.shape[-3:]
            fov = np.array(
                [self.attributes.get(ReconAttributeTags.OLD_FIELD_OF_VIEW) if x != 1 else 1 for x in n_pixels])
        return fov

    def __init__(self, raw_data, ax_1_labels=None,
                 algorithm_id="", field_of_view=None,
                 attributes=None, hdf5_sub_name=None, ax1_meaning=None):
        # Ax1 labels = e.g. Wavelength
        if ax1_meaning is None:
            ax1_meaning = self.get_ax1_label_meaning()

        # Quick bit of validation
        if ax_1_labels is not None and ax1_meaning is not None:
            if raw_data.shape[1] != len(ax_1_labels):
                raise ValueError("Axis 1 labels must match raw data size.")

        if field_of_view is None:
            field_of_view = [1, 1, 1]

        if type(field_of_view[0]) not in [tuple, list, np.ndarray]:
            field_of_view = [(-x / 2, x / 2) if x is not None else (0, 0) for x in field_of_view]

        xs = [np.linspace(x, y, N) for (x, y), N in zip(field_of_view, raw_data.shape[-3:][::-1])]

        dims = ["frames", "z", "y", "x"]
        coords = {"frames": np.arange(raw_data.shape[0]),
                  "x": xs[0],
                  "y": xs[1],
                  "z": xs[2]
                  }

        if not self.ax_0_exists():
            del coords["frames"]
            dims = dims[1:]

        if ax1_meaning is not None:
            dims.insert(1, ax1_meaning)
            coords[ax1_meaning] = ax_1_labels
        else:
            # If there isn't really an axis 1 (e.g. for delta so2).
            coords[ax1_meaning] = ax_1_labels[0]

        DataSequence.__init__(self, raw_data, dims, coords, attributes,
                              hdf5_sub_name, algorithm_id)



#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
Data examples
====================

PATATO data - an interface for exemplar data and basic simulations to illustrate the core features of PATATO.
"""

from .simulated_datasets import get_basic_p0, generate_basic_simulation
from .get_example_datasets import get_msot_time_series_example, get_ithera_msot_time_series_example, \
    get_msot_phantom_example

__all_exports = [get_msot_time_series_example, get_ithera_msot_time_series_example,
                 get_msot_phantom_example, get_basic_p0, generate_basic_simulation]

for e in __all_exports:
    e.__module__ = __name__

__all__ = [e.__name__ for e in __all_exports]



#  Copyright (c) Thomas Else 2023.
#  License: MIT

from typing import Sequence

import numpy as np
from .. import ReconstructionAlgorithm

# Add a loading bar
from tqdm.auto import tqdm


class SlowBackprojection(ReconstructionAlgorithm):
    """
    Slow example backprojection.
    """

    def reconstruct(self, time_series: np.ndarray,
                    fs: float,
                    geometry: np.ndarray, n_pixels: Sequence[int],
                    field_of_view: Sequence[float],
                    speed_of_sound: float,
                    **kwargs) -> np.ndarray:
        """

        Parameters
        ----------
        time_series: array_like
            Photoacoustic time series data in a numpy array. Shape: (..., n_detectors, n_time_samples)
        fs: float
            Time series sampling frequency (Hz).
        geometry: array_like
            The detector geometry. Shape: (n_detectors, 3)
        n_pixels: tuple of int
            Tuple of length 3, (nx, ny, nz)
        field_of_view: tuple of float
            Tuple of length 3, (lx, ly, lz) - the size of the reconstruction volume.
        speed_of_sound: float
            Speed of sound (m/s).
        kwargs
            Extra parameters (optional), useful for advanced algorithms (e.g. multi speed of sound etc.).

        Returns
        -------
        array_like
            The reconstructed image.

        """
        print("Running batch of delay and sum reconstruction code.")

        # Get useful parameters:
        dl = speed_of_sound / fs

        # Reshape frames so that we can loop through to reconstruct
        original_shape = time_series.shape[:-2]
        frames = int(np.product(original_shape))
        signal = time_series.reshape((frames,) + time_series.shape[-2:])

        xs, ys, zs = [
            np.linspace(-field_of_view[i] / 2, field_of_view[i] / 2, n_pixels[i]) if n_pixels[i] != 1 else np.array(
                [0.]) for i in range(3)]
        Z, Y, X = np.meshgrid(zs, ys, xs, indexing='ij')

        # Note that the reconstructions are stored in memory in the order z, y, x (i.e. the x axis is the fastest
        # changing in memory)
        output = np.zeros((frames,) + tuple(n_pixels)[::-1])

        for n_frame in tqdm(range(frames), desc="Looping through frames", position=0):
            for n_detector in tqdm(range(signal.shape[-2]), desc="Looping through detectors", position=1, leave=False):
                detx, dety, detz = geometry[n_detector]
                d = (np.sqrt((detx - X) ** 2 + (dety - Y) ** 2 + (detz - Z) ** 2) / dl).astype(np.int32)
                output[n_frame] += signal[n_frame, n_detector, d]
        return output.reshape(original_shape + tuple(n_pixels)[::-1])

    @staticmethod
    def get_algorithm_name() -> str:
        """

        Returns
        -------
        str
            Algorithm name.
        """
        return "Slow Backprojection"



#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
``patato-convert-simpa``: A script to convert simpa HDF5 output into the PATATO hdf5 format.
"""

import argparse
from distutils.util import strtobool

from ..io.hdf5_converter import convert_simpa


def main():
    parser = argparse.ArgumentParser(
        usage="%(prog)s [-hv] input output",
        description="Convert Simpa Data into a standard hdf5 format. ."
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version=f"{parser.prog} version 0.1"
    )
    parser.add_argument('input', nargs=1, type=str, help="Simpa Study Folder")
    parser.add_argument('output', nargs=1, help="Output Folder")
    parser.add_argument('-n', '--name', type=strtobool, default=False, help="Keep the user defined scan name and do "
                                                                            "not use the iThera defined generic name.")
    args = parser.parse_args()

    convert_simpa(args.input[0], args.output[0], args.name)



"""
This defines the data structure for unmixed datasets.
"""

#  Copyright (c) Thomas Else 2023.
#  License: MIT

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    import h5py
    import numpy as np

from ...core.image_structures.image_sequence import ImageSequence
from ...io.attribute_tags import HDF5Tags


class UnmixedData(ImageSequence):
    """
    UnmixedData stores unmixed datasets.
    """

    save_output = True
    @staticmethod
    def is_single_instance():
        return False

    @staticmethod
    def get_ax1_label_meaning():
        return "SPECTRA"

    def get_hdf5_group_name(self):
        return HDF5Tags.UNMIXED

    @staticmethod
    def get_ax1_labels_from_hdf5(dataset: "h5py.Dataset", file: "h5py.File") -> Optional["np.ndarray"]:
        return dataset.attrs["SPECTRA"]

    @property
    def spectra(self):
        return self.ax_1_labels



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import argparse

import h5py
import matplotlib.pyplot as plt
import numpy as np


def init_argparse():
    parser = argparse.ArgumentParser(description="Import iThera recons.")
    parser.add_argument('input', type=str, help="Data File")
    parser.add_argument('binary', type=str, help="iThera Binary")
    parser.add_argument("description", type=str, help="Recon Description")
    parser.add_argument("fov", type=float, help="Field of View (metres) -e.g.0.025")
    parser.add_argument("-t", "--transpose", type=bool, help="Transpose Images", default=False)
    parser.add_argument("-r", "--reverse", type=int, nargs="*", help="Reverse Axes", default=[])
    return parser


def main():
    p = init_argparse()
    args = p.parse_args()

    file = h5py.File(args.input, "r+")

    data = np.fromfile(args.binary, np.double)
    shape = file["raw_data"].shape[:-2]
    nx = int(np.sqrt(data.shape[0] // np.product(shape)))
    data = data.reshape(file["raw_data"].shape[:-2] + (nx, nx))

    if args.transpose:
        data = np.swapaxes(data, -1, -2)
    for r in args.reverse:
        if r == 0:
            data = data[:, :, ::-1, :]
        elif r == 1:
            data = data[:, :, :, ::-1]

    plt.imshow(data[0, 0])
    plt.show()

    if input("Is this the correct orientation?") == "Y":
        recs = file.require_group("recons")
        recs_g = recs.require_group(args.description)
        if "0" in recs_g:
            del recs_g["0"]
        rec = recs_g.create_dataset("0", data=data)
        rec.attrs["RECON_FOV"] = args.fov
    file.close()



#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
``patato-unmix`` is a command line tool for unmixing MSOT data.
"""

import argparse
import glob
import json
import os.path
from os.path import join

import numpy as np
from .. import PAData
from ..unmixing import get_default_unmixing_preset
from .. import read_unmixing_preset
from ..utils import sort_key
from ..utils.pipeline import run_pipeline


def init_argparse():
    parser = argparse.ArgumentParser(description="Process MSOT Data.")
    parser.add_argument('input', type=str, help="Data File")
    parser.add_argument('-w', '--wavelengths', nargs="*", type=float, default=[])
    parser.add_argument('-i', '--wavelengthindices', nargs="*", type=int, default=[])
    parser.add_argument('-p', '--preset', type=str, help="Preset File",
                        default=None)
    parser.add_argument('-c', '--clear', type=bool, help="Clear Old",
                        default=False)
    parser.add_argument('-f', '--filter', type=str, help="Scan Name Filter",
                        default=None)
    return parser


def main():
    p = init_argparse()
    args = p.parse_args()

    preset = args.preset
    if preset is None:
        preset = get_default_unmixing_preset()

    DATA_FOLDER = args.input
    print(DATA_FOLDER)
    # Load json
    with open(preset) as json_file:
        settings = json.load(json_file)

    if os.path.isfile(DATA_FOLDER):
        data_files = [DATA_FOLDER]
    else:
        data_files = sorted(glob.glob(join(DATA_FOLDER, "**", "*.hdf5"), recursive=True), key=sort_key)

    for data_file in data_files:
        print("Processing", data_file)
        data = PAData.from_hdf5(data_file, "r+")

        if np.any(data.get_wavelengths() == 0.):
            print("Strange, some wavelengths = 0 - investigate.", data_file)
            continue

        if not data.get_recon_types():
            print(f"{data_file} has no reconstructions.")

        pipeline = read_unmixing_preset(settings, data)
        if args.filter is not None:
            if str.lower(args.filter) not in str.lower(data.get_scan_name()):
                print(data.get_scan_name(), "skipped")
                continue

        if args.clear:
            # Clear old processing data and processed
            data.delete_recons(recon_groups=["so2", "unmixed", "thb"])

        reconstructions = data.get_scan_reconstructions()
        for recon in data.get_scan_reconstructions():
            current_data = reconstructions[recon]
            run_pipeline(pipeline, current_data, data, n_batch=-1)


if __name__ == "__main__":
    main()



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import logging
from collections import deque
from typing import Tuple

from ..io.hdf.hdf5_interface import HDF5Writer
from ..io.msot_data import PAData
from ..processing.processing_algorithm import ProcessingAlgorithm, ProcessingResult
from typing_extensions import Deque


def run_batch(start_algorithm, input_data, pa_data, start_dict=None):
    # Set up a queue for running datasets, starting with the input data and the first processing algorithm:
    if start_dict is None:
        start_dict = {}
    run_queue: Deque[Tuple[ProcessingAlgorithm, ProcessingResult, dict]] = deque([(start_algorithm, input_data,
                                                                                     start_dict)])

    # List of extracted results.
    results = []
    additional_results = []

    # Keeping popping off the left of the queue until its empty:
    while run_queue:
        # Get the step, input data and the full pa_data
        algorithm, input_data, extra_args = run_queue.popleft()

        logging.info(f"Running: pipeline step {algorithm.__class__.__name__}.")

        # Run the pipeline step.
        run_result = algorithm.run(input_data, pa_data, **extra_args)

        # Take the algorithm's result and add to our output.
        if run_result is not None:
            output, new_extra_args, additional_data = run_result

            # Keep track of results that need to be saved
            if output.save_output:
                results.append(output)

            # Keep track of side-results
            if additional_data is not None:
                for d in additional_data:
                    if d.save_output:
                        additional_results.append(d)

            # Add algorithm's children to end of queue.
            for c in algorithm.children:
                run_queue.append((c, output, new_extra_args))

    return results, additional_results


def run_pipeline(start: ProcessingAlgorithm, input_data, pa_data: PAData, n_batch=-1, save_results=True,
                 output_file=None, **kwargs):
    # Run through the tree that is defined by the start element.
    logging.debug(pa_data.shape)

    # Reshape pa_data to enable the algorithms to run. TODO: see if this is actually necessary.
    if not pa_data.shape:
        pa_data = pa_data[None]
        input_data = input_data[None]

    # Run the datasets in batches. -1 means run all datasets at the same time.
    if n_batch == -1:
        n_batch = pa_data.shape[0]

    logging.info(f"Running {n_batch} batches.")

    # Data that is produced by the algorithm:
    results = []
    additional_results = []

    for i in range(0, input_data.shape[0], n_batch):
        # Loop through batches and run algorithms.
        logging.info(f"Running batch {i//n_batch + 1} of {(input_data.shape[0]-1)//n_batch + 1}")

        # Get the input datasets:
        end_step = min(input_data.shape[0], i + n_batch)
        input_data_batch = input_data[i: end_step]
        pa_data_batch = pa_data[i:end_step]

        # Run the algorithms on the batches.
        batch_results, batch_additional_results = run_batch(start, input_data_batch, pa_data_batch, start_dict=kwargs)

        # Concatenate all the resulting datasets:
        if not results:
            results = batch_results
        else:
            results = [r + b for r, b in zip(results, batch_results)]

        # Concatenate all the side-results.
        if batch_additional_results is not None:
            if not additional_results:
                additional_results = batch_additional_results
            else:
                additional_results = [r + b for r, b in zip(additional_results, batch_additional_results)]

    # Get the output data writer
    if output_file is not None:
        output_data_writer = HDF5Writer(output_file)
    else:
        output_data_writer = pa_data.scan_writer

    # Save the results.
    if save_results:
        for r in results:
            r.save(output_data_writer)
        for r in additional_results:
            r.save(output_data_writer)

    return results, additional_results



#  Copyright (c) Thomas Else 2023.
#  License: MIT

class PreprocessingAttributeTags:
    HIGH_PASS_FILTER = "FILTER_HIGH_PASS"
    LOW_PASS_FILTER = "FILTER_LOW_PASS"
    TIME_INTERPOLATION = "INTERPOLATE_TIME"
    DETECTOR_INTERPOLATION = "INTERPOLATE_DETECTORS"
    IMPULSE_RESPONSE = "IRF"
    HILBERT_TRANSFORM = "HILBERT_TRANSFORM"
    ENVELOPE_DETECTION = "ENVELOPE_DETECTION"
    WINDOW_SIZE = "WINDOW_SIZE"
    PROCESSING_ALGORITHM = "PREPROCESSING_ALGORITHM"
    COUPLANT_CORRECTION = "COUPLANT_CORRECTION"
    COUPLANT_PATH_LENGTH = "COUPLANT_PATH_LENGTH"


class ReconAttributeTags:
    ADDITIONAL_PARAMETERS = "RECONSTRUCTION_PARAMS"
    X_FIELD_OF_VIEW = "RECONSTRUCTION_FIELD_OF_VIEW_X"
    Y_FIELD_OF_VIEW = "RECONSTRUCTION_FIELD_OF_VIEW_Y"
    Z_FIELD_OF_VIEW = "RECONSTRUCTION_FIELD_OF_VIEW_Z"
    X_NUMBER_OF_PIXELS = "RECONSTRUCTION_NX"
    Y_NUMBER_OF_PIXELS = "RECONSTRUCTION_NY"
    Z_NUMBER_OF_PIXELS = "RECONSTRUCTION_NZ"
    RECONSTRUCTION_ALGORITHM = "RECONSTRUCTION_ALGORITHM"
    OLD_FIELD_OF_VIEW = "RECON_FOV"
    OLD_RECON_NX = "RECON_NX"


class UnmixingAttributeTags:
    RESOLUTION_REDUCE = "RESOLUTION_REDUCE"
    WAVELENGTH_RANGE = "WAVELENGTH_RANGE"
    WAVELENGTH_INDICES = "WAVELENGTH_INDICES"
    UNMIXING_WAVELENGTHS = "WAVELENGTHS"
    SPECTRA = "SPECTRA"
    COMPUTE_SO2 = "SO2"
    SUFFIX = "SUFFIX"
    HAEMOGLOBIN = "Hb"
    OXYHAEMOGLOBIN = "HbO2"
    MELANIN = "Melanin"
    ICG = "ICG"


class ROITags:
    Z_POSITION = "z"
    REPETITION = "repetition"
    ROI_NAME = "class"
    ROI_POSITION = "position"
    RUN = "run"
    GENERATED_ROI = "generated"


class GCAttributeTags:
    STEPS = "steps"
    BUFFER = "buffer"
    SKIP_START = "skip_start"


class HDF5Tags:
    POWER = "POWER"
    OVERALL_CORR = "OverallCorrectionFactor"
    RAW_DATA = "raw_data"
    RECONSTRUCTION = "recons"
    UNMIXED = "unmixed"
    SO2 = "so2"
    THB = "thb"
    SPEED_OF_SOUND = "speedofsound"
    SAMPLING_FREQ = "fs"
    SCAN_GEOMETRY = "GEOMETRY"
    WAVELENGTH = "wavelengths"
    IMPULSE_RESPONSE = "irf"
    Z_POSITION = "Z-POS"
    REPETITION = "REPETITION"
    REGIONS_OF_INTEREST = "rois"
    RUN = "RUN"
    DELTA_SO2 = "dso2"
    BASELINE_SO2 = "baseline_so2"
    BASELINE_SO2_STANDARD_DEVIATION = "baseline_so2_sigma"
    TIMESTAMP = "timestamp"
    TEMPERATURE = "TEMPERATURE"
    ULTRASOUND_FRAME_OFFSET = "ultraSound-frame-offset"
    DATE = "date"
    ORIGINAL_NAME = "original_name"
    SCAN_COMMENT = "comment"
    WATER_ABSORPTION_COEFF = "water-absorption-coefficients"
    WATER_PATHLENGTH = "pathlength"
    ULTRASOUND = "ultrasound"
    ULTRASOUND_FIELD_OF_VIEW = "fov"
    SCAN_NAME = "name"
    SEGMENTATION = "seg"
    DELTA_ICG = "dicg"
    BASELINE_ICG = "baseline_icg"
    BASELINE_ICG_SIGMA = "baseline_icg_sigma"



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import tempfile
import unittest

import h5py
import numpy as np

from patato import PAData


class TestHDF5Load(unittest.TestCase):
    def setUp(self) -> None:
        tf = tempfile.TemporaryFile()
        self.file = h5py.File(tf, "a")

        # Add some raw_data
        raw_data = np.zeros((1, 11, 256, 2030))
        self.file.create_dataset("raw_data", data=raw_data)

        # Add some wavelengths
        self.file.create_dataset("wavelengths", data=np.linspace(700, 900, 11))

        # Impulse response
        self.file.create_dataset("irf", data=np.ones(2030))

        # GEOMETRY
        self.file.create_dataset("GEOMETRY", data=np.ones((256, 3)))

        # Reconstructions
        self.file.require_group("recons")
        self.file.require_group("test_recon")
        self.file.create_dataset("recons/test_recon/0", data=np.ones((1, 11, 333, 333, 1)))
        self.file["recons/test_recon/0"].attrs["RECONSTRUCTION_FIELD_OF_VIEW_X"] = 0.025
        self.file["recons/test_recon/0"].attrs["RECONSTRUCTION_FIELD_OF_VIEW_Y"] = 0.025
        self.file["recons/test_recon/0"].attrs["RECONSTRUCTION_FIELD_OF_VIEW_Z"] = 0.
        self.file["recons/test_recon/0"].attrs["RECONSTRUCTION_NX"] = 333
        self.file["recons/test_recon/0"].attrs["RECONSTRUCTION_NY"] = 333
        self.file["recons/test_recon/0"].attrs["RECONSTRUCTION_NZ"] = 1
        self.pa_data = PAData.from_hdf5(self.file)

    def test_from_hdf5_dataset(self):
        r = self.pa_data.get_scan_reconstructions()
        self.assertEqual(len(r), 1)
        self.assertEqual(list(r.keys())[0], ("test_recon", "0"))
        self.assertEqual(self.pa_data.get_scan_reconstructions()[("test_recon", "0")].get_ax1_label_meaning(),
                         "wavelengths")

    def test_slicing(self):
        data = self.pa_data.get_scan_reconstructions()["test_recon", "0"]
        self.assertEqual(data.shape, (1, 11, 333, 333, 1))
        self.assertEqual(data[0].shape, (11, 333, 333, 1))
        self.assertEqual(data[:, 0].shape, (1, 333, 333, 1))

        self.assertTrue(np.all(data[0].ax_1_labels == np.linspace(700, 900, 11)))
        self.assertEqual(data[0, 0].ax_1_labels.size, 1)
        for i in range(11):
            self.assertEqual(data[0, i].ax_1_labels.item(), np.linspace(700, 900, 11)[i])
        self.assertEqual(data[0].cmap, "bone")
        self.assertEqual(data[0].two_dims(), ("y", "z"))

        with self.assertRaises(IndexError):
            print(data[1])
            print(data[-2])
        #



#  Copyright (c) Thomas Else 2023.
#  License: MIT

from __future__ import annotations

from typing import List
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .rois.roi_type import ROI

import numpy as np
from shapely.geometry import Polygon, Point, MultiPolygon


def to_binary_mask(vertices, min_x, max_x, nx, min_y, max_y, ny):
    from matplotlib import path
    if type(vertices) == np.ndarray or type(vertices) == list:
        vert_path = path.Path(vertices, closed=False)
    else:
        vert_path = vertices
    xs = np.linspace(min_x, max_x, nx)
    ys = np.linspace(min_y, max_y, ny)
    X, Y = np.meshgrid(xs, ys)
    points = np.array([X.flatten(), Y.flatten()]).T
    if type(vert_path) in [Polygon, MultiPolygon]:
        points = [Point(r) for r in points]
        mask = np.array([vert_path.contains(r) for r in points])#.reshape(X.shape)
    else:
        mask = vert_path.contains_points(points)
    return mask.reshape(X.shape)


def get_polygon_mask(p, fov_x, fov_y, nx, ny):
    if type(fov_x) not in [list, tuple, np.ndarray]:
        fov_x = [-fov_x/2, fov_x/2]
        fov_y = [-fov_y/2, fov_y/2]
    if type(p) == Polygon:
        mask = to_binary_mask(np.array(p.exterior.coords.xy).T, fov_x[0], fov_x[1], nx,
                              fov_y[0], fov_y[1],
                              ny)
        for interior in p.interiors:
            mask &= ~to_binary_mask(np.array(interior.coords.xy).T, fov_x[0], fov_x[1], nx,
                                    fov_y[0], fov_y[1],
                                    ny)
    elif type(p) == MultiPolygon:
        geoms = list(p.geoms)
        mask = get_polygon_mask(geoms[0], fov_x, fov_y, nx, ny)
        for g in geoms[1:]:
            mask |= get_polygon_mask(g, fov_x, fov_y, nx, ny)
    else:
        print("WARNING: something strange happening...")
        mask = np.zeros((nx, ny))
    return mask


def generate_mask(vertices, fov_x, nx, fov_y=None, ny=None):
    if fov_y is None:
        fov_y = fov_x
    if ny is None:
        ny = nx
    if type(fov_x) not in [list, tuple, np.ndarray]:
        fov_x = [-fov_x/2, fov_x/2]
        fov_y = [-fov_y/2, fov_y/2]
    if type(vertices) in [Polygon, MultiPolygon]:
        mask = get_polygon_mask(vertices, fov_x, fov_y, nx, ny)
    else:
        mask = to_binary_mask(vertices, fov_x[0], fov_x[1], nx, fov_y[0], fov_y[1], ny)

    return mask


def interpolate_rois(rois: List["ROI"], z_positions):
    if len(rois) <= 1 or all([rois[0].z == r.z for r in rois]):
        return []
    from itk import MorphologicalContourInterpolator, image_view_from_array, array_view_from_image
    from .rois.roi_type import ROI
    buffer = 10

    indices = [r.ax0_index[0] for r in rois]

    min_index = min(indices)
    max_index = max(indices)

    zs = z_positions[min_index:max_index + 1, 0]
    points = [r.points for r in rois]
    x_0 = np.min([np.min(p, axis=0) for p in points], axis=0)
    x_1 = np.max([np.max(p, axis=0) for p in points], axis=0)
    dx = np.min((x_1 - x_0) / 200)

    nx = ((x_1 - x_0) / dx).astype(np.int32)
    minx = x_0 - buffer * dx
    maxx = x_1 + (buffer - 1) * dx
    nx = nx + 20

    mask = np.zeros((zs.shape[0], nx[1], nx[0]), np.uint16)

    for r, p in zip(rois, points):
        mask[r.ax0_index - min_index] = to_binary_mask(p, minx[0], maxx[0], nx[0],
                                                       minx[1], maxx[1], nx[1])
    mask = mask
    im = image_view_from_array(mask)
    interp = MorphologicalContourInterpolator(im)
    np_view = array_view_from_image(interp)
    import cv2 as cv
    shape = [np.squeeze(cv.findContours(i, 1, 2)[0][0].astype(float)) for i in np_view.astype(np.uint8)]

    paths = []
    for j, s in enumerate(shape):
        i = j + min_index
        if i not in indices:
            x, y = s.T

            x *= dx
            y *= -dx
            x += x_0[0] - 10.5 * dx
            y = -x_0[1] - y - nx[1] * dx + 10.5 * dx
            paths.append(ROI(np.array([x, -y]).T, zs[j], rois[0].run,
                             rois[0].repetition, rois[0].roi_class + "~interpolated",
                             rois[0].position, rois[0].generated, [i]))

    return paths



#  Copyright (c) Thomas Else 2023.
#  License: MIT

try:
    import jax
    import jax.numpy as jnp
except ImportError:
    jax = None
    jnp = None
from functools import partial


@partial(jax.jit, static_argnums=(2, 3, 4, 5, 6, 7, 8))
def recon_partial(t, geometry, dl, nx, ny, nz, dx, dy, dz):
    """
    Do delay and sum for a single detector.
    """
    z, y, x = jnp.ogrid[0:nz, 0:ny, 0:nx]
    x = (x - (nx - 1) / 2) * dx
    y = (y - (ny - 1) / 2) * dy
    z = (z - (nz - 1) / 2) * dz
    offsets = (jnp.sqrt((geometry[0] - x) ** 2 +
                        (geometry[1] - y) ** 2 +
                        (geometry[2] - z) ** 2) / dl).astype(jnp.int32)
    return t[offsets]


@partial(jax.jit, static_argnums=(2, 3, 4, 5, 6, 7, 8))
def full_recon(t, geometry, dl, nx, ny, nz, dx, dy, dz):
    """
    Do delay and sum for all detectors.
    """
    all_times = jax.vmap(recon_partial, in_axes=(0, 0,) + (None, ) * 7, out_axes=0)
    return jnp.sum(all_times(t, geometry, dl, nx, ny, nz, dx, dy, dz), axis=0)



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import os
import glob
from os.path import join, dirname, split

from ..io.ithera.read_ithera import iTheraMSOT
from ..io.simpa.read_simpa import SimpaImporter

from ..utils import sort_key


def convert_ithera_msot_binary_to_hdf5(input_path: str, output_path: str, update: bool = False,
                                       use_user_defined_scan_name: bool = False):
    """
    This method converts all iThera MSOT files in a folder and exports them to an HDF5 format.

    :param use_user_defined_scan_name: Use a user defined scan name instead of the ithera number.
    :param input_path: A string that represents a directory. In this directory, a bunch of data is located in individual
                        folders following this naming convention: "Scan_N/" with N being an incrementing unique integer.
    :param output_path: The output path is a string representing a directory path pointing to the location where the
                        resulting hdf5 files should be stored.
    :param update: if True, the file will be processed again, even if an hdf5 file containing raw data with the desired
                   output name already exists at the output_path location.
    """
    scan_paths = list(sorted(glob.glob(join(input_path, "**", "Scan_*/"),
                                       recursive=True), key=sort_key))
    if len(scan_paths) < 1:
        print("Found no scans in the input path.")
    errors = []
    for scan_path in scan_paths:
        ithera_defined_scan_name = split(dirname(scan_path))[-1]
        print("WORKING ON", ithera_defined_scan_name)
        try:
            scan = iTheraMSOT(dirname(scan_path))
            if 0 in scan.raw_data.shape:
                print("Scan skipped because scan was not acquired properly", ithera_defined_scan_name)
        except FileNotFoundError as e:
            print(str(e))
            print("Scan skipped, some data missing?", ithera_defined_scan_name)
            errors.append("Scan skipped, some data missing in " + ithera_defined_scan_name)
            continue
        user_defined_scan_name = scan.get_scan_name()

        if use_user_defined_scan_name:
            scan_name = user_defined_scan_name
        else:
            scan_name = ithera_defined_scan_name

        scan_name = join(dirname(dirname(scan_path))[len(input_path) + 1:], scan_name)
        if os.path.exists(join(output_path, scan_name + ".hdf5")) and not update:
            continue
        print("SAVING AS", join(output_path, scan_name + ".hdf5"))

        scan.save_to_hdf5(join(output_path, scan_name + ".hdf5"))
        print("Saved to HDF5.")
    print("\n".join(errors))


def convert_simpa(input_path: str, output_path: str,
                  use_user_defined_scan_name: bool = False):
    """
    This method converts simpa files to HDF5 format.

    Parameters
    ----------
    input_path
    output_path
    update
    use_user_defined_scan_name
    slice_n
    """
    scan_paths = list(sorted(glob.glob(join(input_path, "**", "*.hdf5"),
                                       recursive=True), key=sort_key))
    if len(scan_paths) < 1:
        print("Found no scans in the input path.")

    i = 1
    for scan_path in scan_paths:
        print("WORKING ON", scan_path)

        scan = SimpaImporter(scan_path)
        user_defined_scan_name = scan.get_scan_name()

        if use_user_defined_scan_name:
            scan_name = user_defined_scan_name
        else:
            scan_name = f"Scan_{str(i)}"
            i += 1

        scan_name = join(dirname(dirname(scan_path))[len(input_path) + 1:], scan_name)
        print(scan_name)
        print("SAVING AS", join(output_path, scan_name + ".hdf5"))

        scan.save_to_hdf5(join(output_path, scan_name + ".hdf5"))
        print("Saved to HDF5.")



"""
Defines a datatype for datasets that have been processed to only have one value per scan per pixel.
"""

#  Copyright (c) Thomas Else 2023.
#  License: MIT

from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    import h5py

from ...core.image_structures.image_sequence import ImageSequence


class SingleImage(ImageSequence):
    """
    SingleImage is the datastructure for images like delta sO2..
    """

    @staticmethod
    def get_ax1_label_meaning():
        return None
    save_output = True

    def get_hdf5_group_name(self):
        return self.ax_1_labels.item()

    @staticmethod
    def get_ax1_labels_from_hdf5(dataset: "h5py.Dataset", file: "h5py.File") -> List[str]:
        return [dataset.name.split("/")[-3]]

    @staticmethod
    def is_single_instance():
        return False

    @property
    def ax_0_labels(self):
        return None

    @staticmethod
    def ax_0_exists():
        return False



#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
Useful utilities module
==============================

This module provides miscellaneous functions and classes that are useful for PATATO.
They are largely independent of the other modules and minimally documented.
"""

from typing import Optional

import numpy as np
import pandas as pd

from ..utils.process_study import get_hdf5_files

import matplotlib.transforms as transforms
import matplotlib

def add_axis_label(ax, label, manual_x=0, manual_y=0, font_size="large"):
    fig = ax.get_figure()
    fig.canvas.draw()
    bbox = ax.get_tightbbox(fig.canvas.get_renderer())
    
    x, y = fig.transSubfigure.inverted().transform([bbox.x0, bbox.y1])
    transform = fig.transSubfigure
    if ax.get_subplotspec().is_first_col():
        transform = transforms.blended_transform_factory(ax.get_figure().transSubfigure, transform)
        x = 0
        transform += transforms.ScaledTranslation(matplotlib.rcParams['figure.constrained_layout.w_pad'], 0, 
            ax.get_figure().dpi_scale_trans)
    manual_x *= matplotlib.font_manager.font_scalings.get(font_size, 1)
    manual_y *= matplotlib.font_manager.font_scalings.get(font_size, 1)
    if manual_x != 0 or manual_y !=0:
        transform += transforms.ScaledTranslation(manual_x, manual_y, 
            ax.get_figure().dpi_scale_trans)
    return fig.text(x, y, label, fontsize=font_size, fontweight="bold", 
                    va="top", ha="left", transform=transform)
                    
def add_subfigure_label(subfig, ax, label, manual_x=0, manual_y=0, font_size="large"):
    manual_x *= matplotlib.font_manager.font_scalings.get(font_size, 1)
    manual_y *= matplotlib.font_manager.font_scalings.get(font_size, 1)
    t = subfig.suptitle(label, ha="left", va="top", fontweight="bold", fontsize=font_size)
    subfig.canvas.draw()
    t.set_x(0)
    transform = subfig.transSubfigure + transforms.ScaledTranslation(matplotlib.rcParams['figure.constrained_layout.w_pad'], 
    0, 
    subfig.get_figure().dpi_scale_trans)
    if manual_x != 0 or manual_y !=0:
        transform += transforms.ScaledTranslation(manual_x, manual_y, 
            ax.get_figure().dpi_scale_trans)
    t.set_transform(transform)
    return t

def linear_regression(x, y, constant=True, x_predict=None):
    import statsmodels.api as sm 
    if constant:
        x = sm.add_constant(x)
    model = sm.OLS(y, x)
    result = model.fit()
    # prediction = result.get_prediction()
    if x_predict is None:
        x_predict = x
    else:
        x_predict = sm.add_constant(x_predict)
    prediction = result.get_prediction(x_predict)
    return prediction, result
       
    

def process_scan_name(template: str, scan_name: str) -> dict:
    """
    Process the scan name using the simple template specified.

    The template should contain something like the following:

    >>> "<date>_<initials><earmark><mouseid>_Day<timepoint>_<scantype>"

    For advanced use, note that this is converted into a regular expression,
    so certain elements of that syntax can be used in the template.

    Parameters
    ----------
    template
    scan_name

    Returns
    -------

    """
    import re
    regex_codes = {"Date": r"([0-9]*)",
                   "Initials": r"([A-z]{2,3})",
                   "EarMark": r"(NM|1L|1R|2L|2R|1L1R|1R1L|1RL|1B|IB|IL|IR)",
                   "MouseID": r"([0-9]{1,6})",
                   "ScanType": r"([A-z|0-9|_|\+]+)?",
                   "Timepoint": r"([0-9]+)"}

    # template = template.replace("(", "(?:")
    for k, code in regex_codes.items():
        code = f"(?P<{k}>" + code[1:]
        template = template.replace(f"<{k}>", code)

    p = re.compile(template)
    try:
        d = p.match(scan_name).groupdict()
        for k, v in d.items():
            if type(v) == str:
                d[k] = v.upper()
        return d
    except AttributeError:
        print(f"Unable to match template to scan name: {scan_name}.")
        return {}


def invert_dictionary_tolist(mapping):
    return {mouse: date for date, mice in mapping.items() for mouse in mice}


def extract_data_tables(datafolder: str, name_template: str,
                        analyse_rois: list, metrics=None,
                        start_days=None,
                        group_info: Optional[dict] = None,
                        reconstruction_name=None,
                        analyse_scan_types=None,
                        just_summary=True,
                        roi_kwargs=None,
                        apply_function=None,
                        filter_name="",
                        more_details=None,
                        roi_source_type=None,
                        return_masks=False
                        ):
    if start_days is None:
        start_days = {}
    if group_info is None:
        group_info = {}
    if metrics is None:
        metrics = ["thb", "so2"]

    start_date_map = {timepoint: invert_dictionary_tolist(mapping) for timepoint, mapping in start_days.items()}
    group_info = {timepoint: invert_dictionary_tolist(mapping) for timepoint, mapping in group_info.items()}

    images = []
    tables = []

    # Share regions of interest between adjacent scans.
    datasets = list(get_hdf5_files(datafolder, filter_name=filter_name))
    dataset_details = [(process_scan_name(name_template, data.get_scan_name()), data) for _, data in datasets]

    # Generate a dictionary to lookup all the scans for each scan session. Using the MouseID and Timepoint as an ID.
    scan_map = {}
    for details, data in dataset_details:
        scan_id = (details.get("MouseID", None), details.get("Timepoint", None))
        if scan_id not in scan_map:
            scan_map[scan_id] = {}
        scan_map[scan_id][details.get("ScanType")] = data

    if roi_source_type is not None:
        for scan_id in scan_map:
            if roi_source_type not in scan_map[scan_id]:
                continue
            for scan_type in scan_map[scan_id]:
                scan_map[scan_id][scan_type].external_roi_interface = scan_map[scan_id][roi_source_type]

    # Loop through all datasets and extract data.
    for f, data in datasets:
        print(f, data.get_scan_name())
        # Set the default reconstruction method.
        data.set_default_recon(reconstruction_name)

        # Extract useful information from the scan name
        scan_name = str.strip(data.get_scan_name())
        details = process_scan_name(name_template, scan_name)

        if analyse_scan_types is not None:
            if details.get("ScanType", None) not in analyse_scan_types:
                continue

        if not details:
            continue
        details["File"] = f
        # Mouse id - must be set for this analysis code.
        mouse_id = int(details["MouseID"])

        # Get the scan date.
        date = data.get_scan_datetime()

        # Extract details (e.g treatment, cell line etc)
        for detail, mouse_mapping in group_info.items():
            details[detail] = mouse_mapping[mouse_id]

        # Extract the time data (e.g. time since dosing started, time since implantation)
        for time_detail, mouse_mapping in start_date_map.items():
            details[time_detail] = (date - mouse_mapping[mouse_id]).days + 1

        if more_details is not None:
            for fn in more_details:
                for k, v in fn(data).items():
                    details[k] = v

        if not data.get_rois():
            continue
        else:
            measurements = data.summary_measurements(metrics=metrics,
                                                     include_rois=analyse_rois,
                                                     roi_kwargs=roi_kwargs,
                                                     just_summary=just_summary,
                                                     return_masks=return_masks)
            for d in details:
                measurements[d] = details[d]
            measurements["Date"] = data.get_scan_datetime()
            if apply_function is not None:
                measurements = apply_function(measurements)
            tables.append(measurements)

    df = tables[0].append(tables[1:]).reset_index()

    df["Radius"] = np.sqrt(df["Area"] / np.pi) * 75 * 3e-3

    df["Volume"] = df["Radius"] ** 3 * 4 * np.pi / 3

    df["Date"] = pd.to_datetime(df["Date"], utc=True, dayfirst=True).dt.tz_localize(None)

    return df, images


def set_matplotlib_defaults(fig_width=91.5, fig_height=89):
    import matplotlib
    matplotlib.rcParams["pdf.fonttype"] = 42
    matplotlib.rcParams["ps.fonttype"] = 42
    matplotlib.rcParams["font.sans-serif"] = "Arial"
    matplotlib.rcParams["figure.dpi"] = 227
    matplotlib.rcParams["figure.figsize"] = (
        fig_width / 25.4, fig_height / 25.4)  # OR 183 mm for double width
    matplotlib.rcParams["font.size"] = 7
    matplotlib.rcParams["axes.spines.top"] = False
    matplotlib.rcParams["axes.spines.right"] = False
    matplotlib.rcParams["savefig.pad_inches"] = 0
    matplotlib.rcParams['figure.subplot.bottom'] = 0.075
    matplotlib.rcParams['figure.subplot.hspace'] = 0.4
    matplotlib.rcParams['figure.subplot.left'] = 0.075
    matplotlib.rcParams['figure.subplot.right'] = 0.97
    matplotlib.rcParams['figure.subplot.top'] = 0.925
    matplotlib.rcParams['figure.subplot.wspace'] = 0.5
    matplotlib.rcParams["figure.titlesize"] = "medium"
    matplotlib.rcParams['axes.titlesize'] = "small"
    matplotlib.rcParams["lines.markersize"] = 3
    matplotlib.rcParams["legend.frameon"] = False



#  Copyright (c) Thomas Else 2023.
#  License: MIT

from functools import partial
from typing import Union, Tuple, Optional, Dict

import numpy as np
from patato.io.attribute_tags import PreprocessingAttributeTags
from scipy.signal.windows import hann

from .processing_algorithm import TimeSeriesProcessingAlgorithm

try:
    import jax.numpy as jnp
    import jax
except ImportError:
    jnp = None
    jax = None

import warnings

from ..core.image_structures.pa_time_data import PATimeSeries

# Specify Array type
Array = np.typing.NDArray

@jax.jit
def subtract_mean(time_series):
    # A function to subtract the mean from a time series
    return time_series - jnp.mean(time_series, axis=-1).reshape(time_series.shape[:-1] + (1,))


@partial(jax.jit, static_argnums=(1,))
def interpolate_detectors(detectors, ndet):
    # Interpolate the detectors to the correct number of detectors
    new_detector_i = jnp.linspace(0, detectors.shape[0] - 1, ndet * detectors.shape[0])
    old_detector_i = jnp.arange(detectors.shape[0])
    interp_detectors = jax.vmap(jnp.interp, in_axes=(None, None, -1), out_axes=-1)
    new_detectors = interp_detectors(new_detector_i, old_detector_i, detectors)
    return new_detectors


@partial(jax.jit, static_argnums=(1, 2))
def partial_interpolate(time_series: Array, nt: int, ndet: int) -> Array:
    # Interpolate the time series to the correct number of time points
    new_detector_i = jnp.linspace(0, time_series.shape[-2] - 1, ndet * time_series.shape[-2])
    old_detector_i = jnp.arange(time_series.shape[-2])
    new_times = jnp.linspace(0, time_series.shape[-1] - 1, nt * time_series.shape[-1] + 1)[:-1]
    old_times = jnp.arange(time_series.shape[-1])

    interp_detectors = jax.vmap(jnp.interp, in_axes=(None, None, -1), out_axes=-1)
    new_time_series = interp_detectors(new_detector_i, old_detector_i, time_series)

    interp_time = jax.vmap(jnp.interp, in_axes=(None, None, 0), out_axes=0)
    new_time_series = interp_time(new_times, old_times, new_time_series)
    return new_time_series


def make_filter(n_samples: int, fs: float, irf: Array, hilbert: bool, lp_filter: Optional[float],
                hp_filter: Optional[float], rise: float = 0.2, n_filter: int = 1024, window: Optional[str] = None):
    """
    Make the filter for the time series

    Parameters
    ----------
    n_samples : int
    fs : float
    irf : Array
    hilbert : bool
    lp_filter: float or None
    hp_filter: float or None
    rise: float
    n_filter : int
    window

    Returns
    -------

    """
    output = np.ones((n_samples,), dtype=np.cdouble)

    # Impulse response correction
    if irf is not None:
        irf_shifted = np.fft.fftshift(irf)
        # Divide by the impulse response to deconvolve
        output *= np.conj(np.fft.fft(irf_shifted)) / np.abs(np.fft.fft(irf_shifted)) ** 2
        # Suppress high frequencies to avoid amplifying noise - apply a window.
        output *= np.fft.fftshift(hann(n_samples))

    # Hilbert Transform
    frequencies = np.fft.fftfreq(n_samples)
    if hilbert:
        # TODO: check this
        # Multiply positive frequencies by
        output *= (1 + np.sign(frequencies)) / 2

    frequencies = np.abs(np.fft.fftfreq(n_filter, 1 / fs))
    filter_output = np.ones_like(frequencies, dtype=np.cdouble)

    if hp_filter is not None:
        filter_output[frequencies < hp_filter * (1 - rise)] = 0
        in_rise = np.logical_and(frequencies > hp_filter * (1 - rise), frequencies < hp_filter)
        filter_output[in_rise] = (frequencies[in_rise] - hp_filter * (1 - rise)) / (hp_filter * rise)

    if lp_filter is not None:
        filter_output[frequencies > lp_filter * (1 + rise)] = 0
        in_rise = np.logical_and(frequencies < lp_filter * (1 + rise), frequencies > lp_filter)
        filter_output[in_rise] = 1 - (frequencies[in_rise] - lp_filter) / (lp_filter * rise)

    time_series = np.fft.ifft(filter_output)

    if window == "hann":
        time_series *= np.fft.fftshift(hann(n_filter))

    filter_time = np.zeros_like(output)
    filter_time[:n_filter // 2] = time_series[:n_filter // 2]
    filter_time[-n_filter // 2:] = time_series[-n_filter // 2:]

    filter_output = np.fft.fft(filter_time)
    output *= filter_output
    return output


class PreProcessor(TimeSeriesProcessingAlgorithm):
    """ Preprocesses MSOT time series data. Uses JAX in the background.
    """

    @staticmethod
    def get_algorithm_name() -> Union[str, None]:
        """
        Get the name of the algorithm.

        Returns
        -------
        str or None
        """
        return "Standard Preprocessor"

    @staticmethod
    def get_hdf5_group_name() -> Union[str, None]:
        """
        Return the name of the group in the HDF5 file

        Returns
        -------
        str or None

        """
        # Return the name of the group in the HDF5 file
        return None

    def __init__(self, time_factor: int = 3, detector_factor: int = 2,
                 irf: bool = True, hilbert: bool = True, lp_filter: Optional[float] = None,
                 hp_filter: Optional[float] = None, filter_window_size: int = 512,
                 window: str = "hann", absolute: Optional[str] = None, universal_backprojection=False):
        # Initialise the preprocessor
        super().__init__()
        self.time_factor = time_factor
        self.detector_factor = detector_factor
        self.hilbert = hilbert
        absolute = "imag" if absolute is None and hilbert else absolute
        self.ubp = universal_backprojection
        self.irf_correct = irf
        self.lp_filter = lp_filter
        self.hp_filter = hp_filter
        self.n_filter = filter_window_size
        self.window = window
        self.absolute = absolute
        self.filter = None

    def pre_compute_filter(self, n_samples: int, fs: float, irf: Array = None):
        """
        Precompute the filter to be applied.

        Parameters
        ----------
        n_samples : int
        fs : float
        irf : Array
        """
        self.filter = jnp.array(make_filter(n_samples, fs, irf, self.hilbert, self.lp_filter, self.hp_filter))

    def _run(self, time_series: Array, detectors: Array, overall_correction_factor, **kwargs):
        """
        Run the preprocessing step on a given time series and detectors. This allows batch processing,
         e.g. if the data doesn't fit into memory.

        Parameters
        ----------
        time_series : Array
        detectors   : Array
        overall_correction_factor: Array
        kwargs    : dict

        Returns
        -------
        tuple of Array, Array
        """
        shape = time_series.shape
        time_series = jnp.array(time_series.reshape((-1,) + shape[-2:]))
        detectors = jnp.array(detectors)

        # Subtract mean
        time_series = subtract_mean(time_series)

        # Apply filters
        time_series_ft = jnp.fft.fft(time_series)

        time_series_ft = time_series_ft * self.filter.reshape((1,) * (time_series.ndim - 1) + (-1,))
        if self.absolute == "imag":
            op = jnp.imag
        else:
            op = jnp.real
        time_series = op(jnp.fft.ifft(time_series_ft))

        # Allow for universal backprojection here.
        if self.ubp:
            time_series -= jnp.gradient(time_series, axis=-1)*jnp.arange(time_series.shape[-1])

        # Interpolate in time domain and detector domain
        if not (self.detector_factor == 1 and self.time_factor == 1):
            full_interpolate = jax.vmap(partial_interpolate, in_axes=(0, None, None), out_axes=0)
            time_series = full_interpolate(time_series,
                                           self.time_factor,
                                           self.detector_factor)
            detectors = interpolate_detectors(detectors, self.detector_factor)
        time_series = time_series.reshape(shape[:-2] + time_series.shape[1:])

        # Apply energy correction factor:
        if overall_correction_factor is not None:
            extend = (slice(None, None),) * overall_correction_factor.ndim + (None, None)
            time_series /= overall_correction_factor[extend]
        else:
            warnings.warn("No energy correction factor applied.")

        return time_series, detectors

    def run(self, time_series, pa_data=None, irf=None, detectors=None, **kwargs) -> Tuple[
        PATimeSeries, Dict, Optional[list]]:
        """
        Run the preprocessing step on a given time series and detectors. This allows batch processing,
        e.g. if the data doesn't fit into memory.

        Parameters
        ----------
        time_series
        pa_data
        irf
        detectors
        kwargs

        Returns
        -------
        tuple of PATimeSeries, dict, list
        """
        from .. import PAT_MAXIMUM_BATCH_SIZE
        # Impulse response
        if irf is None and pa_data is not None:
            irf = pa_data.get_impulse_response()

        # Photoacoustic transducers
        if detectors is None and pa_data is not None:
            detectors = pa_data.get_scan_geometry()

        if pa_data is not None:
            overall_correction_factor = pa_data.get_overall_correction_factor()
        else:
            overall_correction_factor = None

        # Sampling frequency
        fs = time_series.attributes["fs"]

        if self.filter is None:
            self.pre_compute_filter(time_series.shape[-1], fs=fs, irf=irf)

        new_detectors = detectors
        if time_series.shape[0] * time_series.shape[1] > PAT_MAXIMUM_BATCH_SIZE != -1:
            new_timeseries = []
            ts_raw = time_series.raw_data
            shape = ts_raw.shape
            ts_raw = ts_raw.reshape((-1,) + shape[-2:])
            overall_correction_factor = overall_correction_factor.flatten()
            for i in range(0, ts_raw.shape[0], PAT_MAXIMUM_BATCH_SIZE):
                new_ts, new_detectors = self._run(ts_raw[i:i + PAT_MAXIMUM_BATCH_SIZE], detectors,
                                                  overall_correction_factor[i:i + PAT_MAXIMUM_BATCH_SIZE])
                new_timeseries.append(np.asarray(new_ts))
            new_ts = np.concatenate(new_timeseries, axis=0).reshape(shape[:2] + new_timeseries[0].shape[-2:])
        else:
            new_ts, new_detectors = self._run(time_series.raw_data, detectors,
                                              overall_correction_factor)

        # Convert timeseries into an xarray
        attributes = dict(time_series.attributes)
        attributes["fs"] *= self.time_factor
        attributes[PreprocessingAttributeTags.IMPULSE_RESPONSE] = self.irf_correct
        attributes[PreprocessingAttributeTags.PROCESSING_ALGORITHM] = self.get_algorithm_name()
        attributes[PreprocessingAttributeTags.WINDOW_SIZE] = self.window
        attributes[PreprocessingAttributeTags.ENVELOPE_DETECTION] = self.absolute == "abs"
        attributes[PreprocessingAttributeTags.HILBERT_TRANSFORM] = self.hilbert
        attributes[PreprocessingAttributeTags.DETECTOR_INTERPOLATION] = self.detector_factor
        attributes[PreprocessingAttributeTags.TIME_INTERPOLATION] = self.time_factor
        attributes[PreprocessingAttributeTags.LOW_PASS_FILTER] = self.lp_filter
        attributes[PreprocessingAttributeTags.HIGH_PASS_FILTER] = self.hp_filter
        attributes["UniversalBackProjection"] = self.ubp
        attributes["CorrectionFactorApplied"] = overall_correction_factor is not None

        coords = dict(time_series.da.coords)
        coords["detectors"] = np.linspace(0, time_series.shape[-2] - 1,
                                          self.detector_factor * time_series.shape[-2])
        coords["timeseries"] = np.linspace(0, time_series.shape[-1] - 1,
                                           self.time_factor * time_series.shape[-1] + 1)[:-1]

        new_data = PATimeSeries(new_ts, time_series.da.dims, coords, attributes=attributes)
        return new_data, {"geometry": new_detectors}, None



#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
Utility module
=====================

This module provides utility functions for PATATO that are used in several other modules.
"""

from os.path import split
from typing import Tuple


def sort_key(file: str) -> Tuple[int, str]:
    """

    Parameters
    ----------
    file

    Returns
    -------

    """
    try:
        return int(split(file)[-1].split(".")[0].split("_")[-1]), file
    except ValueError:
        return 0, file



#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
Core datastructures
===================

Core datastructure for PATATO. This module provides structures such as Reconstruction, UnmixedData etc., providing
commonly-used functions.
"""



#  Copyright (c) Thomas Else 2023.
#  License: MIT

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional
from typing import Union, List, Any, Tuple

if TYPE_CHECKING:
    from ..io.msot_data import PAData
    from ..core.image_structures.image_sequence import ImageSequence
    from ..core.image_structures.pa_time_data import PATimeSeries


class ProcessingResult(ABC):
    save_output = False

    @abstractmethod
    def __add__(self, other):
        pass

    @property
    def attributes(self):
        return {}

    def __init__(self):
        self.algorithm_id = ""
        self.hdf5_sub_name = ""

    @property
    @abstractmethod
    def raw_data(self):
        pass

    @abstractmethod
    def get_hdf5_group_name(self):
        pass

    @staticmethod
    @abstractmethod
    def is_single_instance():
        pass

    def get_hdf5_sub_name(self):
        return self.hdf5_sub_name

    def save(self, scan_writer):
        scan_writer.add_image(self)


class ProcessingAlgorithm(ABC):
    def __init__(self):
        self.children: List["ProcessingAlgorithm"] = []

    @abstractmethod
    def run(self, input_data: Any,
            pa_data: PAData, **kwargs) -> Optional[Tuple[ProcessingResult, dict, Optional[List[ProcessingResult]]]]:
        pass

    def add_child(self, child: "ProcessingAlgorithm"):
        self.children.append(child)


class TimeSeriesProcessingAlgorithm(ProcessingAlgorithm, ABC):
    def __init__(self):
        ProcessingAlgorithm.__init__(self)

    @abstractmethod
    def run(self, time_series: PATimeSeries,
            pa_data: PAData, **kwargs):
        pass

    @staticmethod
    @abstractmethod
    def get_algorithm_name() -> Union[str, None]:
        pass


class SpatialProcessingAlgorithm(ProcessingAlgorithm, ABC):
    def __init__(self, algorithm_id=""):
        ProcessingAlgorithm.__init__(self)
        self.algorithm_id = algorithm_id

    @abstractmethod
    def run(self, spatial_data: ImageSequence,
            pa_data: PAData, **kwargs):
        pass

    def get_algorithm_id(self) -> Union[str, None]:
        return self.algorithm_id



#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
Reconstruction module
============================

This module provides functions for reconstructing photoacoustic data.
"""

from typing import Sequence

from .backprojection_opencl import OpenCLBackprojection
from .backprojection_reference import ReferenceBackprojection
from .reconstruction_algorithm import ReconstructionAlgorithm
from .numpy_backprojection.recon import SlowBackprojection
from ..io.msot_data import PAData

RECONSTRUCTION_METHODS: Sequence[type(ReconstructionAlgorithm)] = [OpenCLBackprojection,
                                                                   ReferenceBackprojection,
                                                                   SlowBackprojection]

RECONSTRUCTION_NAMES = {x.get_algorithm_name(): x for x in RECONSTRUCTION_METHODS}


def get_default_recon_preset(data: PAData):
    """

    Parameters
    ----------
    data

    Returns
    -------

    """
    import numpy as np
    import os
    try:
        import pyopencl as cl
        CPU = not any(len(platform.get_devices(device_type=cl.device_type.GPU)) > 0 for platform in cl.get_platforms())
    except ImportError:
        cl = None
        CPU = False  # Should this maybe be true?
    root_folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), "recon_presets")
    preset = ""

    if np.isclose(np.sqrt(np.sum(data.get_scan_geometry()[0] ** 2)), 0.0405):
        preset += "backproject_standard"
    else:
        preset += "backproject_clinical"
    if CPU:
        preset += "_CPU"

    # Get axes:
    axes = ""
    for i in range(3):
        if not np.all(np.isclose(data.get_scan_geometry()[:, i],  data.get_scan_geometry()[0, i])):
            axes += "xyz"[i]

    preset += "_" + axes
    return os.path.join(root_folder, preset + ".json")



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import os
from abc import ABC, abstractmethod
from os.path import join
from typing import Union

import numpy as np


# ALL UNITS IN INVERSE CENTIMETRES (some per mole where specified)

class Spectrum(ABC):
    @staticmethod
    @abstractmethod
    def get_name() -> Union[str, None]:
        pass

    @staticmethod
    @abstractmethod
    def get_spectrum(wavelengths: np.ndarray) -> np.ndarray:
        pass


class Haemoglobin(Spectrum):
    # In cm-1/M, divide by 187 for indicative mua (see prahl)
    @staticmethod
    def get_name() -> str:
        return "Hb"

    @staticmethod
    def get_spectrum(wavelengths: np.ndarray) -> np.ndarray:
        from pandas import read_table
        folder = os.path.dirname(os.path.realpath(__file__))
        prahl_spec = read_table(join(folder, "spectra_files", "prahl.txt"))
        return np.interp(wavelengths, prahl_spec["lambda"], prahl_spec["Hb"], np.nan, np.nan)


class OxyHaemoglobin(Spectrum):
    # In cm-1/M, divide by 187 for indicative mua (see prahl)
    @staticmethod
    def get_name() -> str:
        return "HbO2"

    @staticmethod
    def get_spectrum(wavelengths: np.ndarray) -> np.ndarray:
        from pandas import read_table
        folder = os.path.dirname(os.path.realpath(__file__))
        prahl_spec = read_table(join(folder, "spectra_files", "prahl.txt"))
        return np.interp(wavelengths, prahl_spec["lambda"], prahl_spec["Hb02"], np.nan, np.nan)


class IndocyanineGreen(Spectrum):
    @staticmethod
    def get_name() -> str:
        return "ICG"

    @staticmethod
    def get_spectrum(wavelengths: np.ndarray) -> np.ndarray:
        from pandas import read_csv
        folder = os.path.dirname(os.path.realpath(__file__))
        df = read_csv(join(folder, "spectra_files", "ICG.csv"), header=None, sep=";", decimal=",")
        df = df[df[0] < 890]
        return np.interp(wavelengths, df[0], df[1], np.nan, np.nan)


class Melanin(Spectrum):
    # NOTE: Multiply by a volume fraction (~0.01-0.4) to get a realistic epidermis absorption.
    @staticmethod
    def get_name() -> str:
        return "Melanin"

    @staticmethod
    def get_spectrum(wavelengths: np.ndarray) -> np.ndarray:
        if type(wavelengths) != np.ndarray:
            wavelengths = np.array(wavelengths)
        return 1.7e12 * wavelengths ** (-3.48)


class Lipids(Spectrum):
    # From https://omlc.org/spectra/fat/fat.txt
    @staticmethod
    def get_name() -> str:
        return "Lipid"

    @staticmethod
    def get_spectrum(wavelengths: np.ndarray) -> np.ndarray:
        from pandas import read_table
        # print(os.path.realpath(__file__))
        folder = os.path.dirname(os.path.realpath(__file__))
        df = read_table(join(folder, "spectra_files", "lipids.txt"), skiprows=4)
        return np.interp(wavelengths, df["nm"], df["mu_a(1/m)"] / 300, np.nan, np.nan)


class Water(Spectrum):
    @staticmethod
    def get_name() -> str:
        return "H2O"

    @staticmethod
    def get_spectrum(wavelengths: np.ndarray) -> np.ndarray:
        import pandas as pd
        folder = os.path.dirname(os.path.realpath(__file__))
        water_file = join(folder, "spectra_files", "water.txt")
        water_short = pd.read_table(water_file)
        water_file_long = join(folder, "spectra_files", "water_long.txt")
        water_long = pd.read_table(water_file_long)

        water_long = water_long.sort_values("lambda", ignore_index=True)
        # regrid
        min_wavelength = water_short["lambda"].min()
        max_wavelength = water_long["lambda"].max()
        grid_wavelengths = np.arange(min_wavelength, max_wavelength, 5)
        short = np.interp(grid_wavelengths, water_short["lambda"], water_short["absorption"], np.nan, np.nan)
        long = np.interp(grid_wavelengths, water_long["lambda"], water_long["absorption"], np.nan, np.nan)
        absorption_values = np.nanmean(np.array([short, long]), axis=0)
        return np.interp(wavelengths, grid_wavelengths, absorption_values)


SPECTRA = [Haemoglobin, OxyHaemoglobin, IndocyanineGreen, Melanin, Water, Lipids]
SPECTRA_NAMES = {s.get_name(): s for s in SPECTRA}



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import glob
from os.path import join

from ..io.msot_data import PAData
from . import sort_key


def get_hdf5_files(folder, filter_name="", mode="r"):
    for file in sorted(glob.glob(join(folder, "**", "*.hdf5"), recursive=True), key=sort_key):
        pa_data = PAData.from_hdf5(file, mode)
        if filter_name.lower() not in pa_data.get_scan_name().lower() and filter_name.lower() not in file.lower():
            pa_data.scan_reader.close()
            continue
        else:
            yield file, pa_data



#  Copyright (c) Thomas Else 2023.
#  License: MIT

from typing import Tuple, Optional, List, Union

import numpy as np

from ..core.image_structures.reconstruction_image import Reconstruction
from ..core.image_structures.single_image import SingleImage
from ..core.image_structures.single_parameter_data import SingleParameterData
from ..core.image_structures.unmixed_image import UnmixedData
from ..io.attribute_tags import UnmixingAttributeTags, GCAttributeTags
from ..io.msot_data import PAData, HDF5Tags
from ..processing.processing_algorithm import SpatialProcessingAlgorithm, ProcessingResult
from ..unmixing.spectra import Spectrum, SPECTRA_NAMES
from ..utils.time_series_analysis import find_gc_boundaries
from ..utils.rois.roi_type import ROI


class SpectralUnmixer(SpatialProcessingAlgorithm):
    """The spectral unmixer. This takes in reconstruction data and spits out linearly spectrall unmixed data.
    """

    @staticmethod
    def re_grid(reconstruction: np.ndarray, scaling_factor: int):
        """

        Parameters
        ----------
        reconstruction
        scaling_factor

        Returns
        -------

        """
        if scaling_factor == 1:
            return reconstruction
        else:
            n_spatial = 3 if reconstruction.ndim == 5 else 2
            reshape_kernel_size = tuple(min(scaling_factor, size) for size in reconstruction.shape[-n_spatial:])
            extend = (None,) * len(reconstruction.shape[:-n_spatial])
            kernel = np.ones(reshape_kernel_size)[extend]
            kernel /= np.sum(kernel)
            from scipy.signal import convolve
            smoothed = convolve(reconstruction, kernel, mode="same")
            slice_selection = (slice(None, None),) * len(reconstruction.shape[:-n_spatial])
            slice_selection += (slice(None, None, scaling_factor),) * n_spatial
            return smoothed[slice_selection]

    def run(self, reconstruction: Reconstruction, _, **kwargs) -> Tuple[UnmixedData, dict, None]:
        # Select the right wavelengths:
        wavelengths = self.wavelengths
        wavelength_indices = np.where(np.isclose(reconstruction.wavelengths[:, None], wavelengths[None, :]))[0]

        # Get the reconstructed data
        recon_data = reconstruction[:, wavelength_indices].raw_data

        # Change the grid
        recon_data = self.re_grid(recon_data, self.rescaling_factor)

        # Unmix.
        unmixed = np.einsum('ij...,jk->ik...', recon_data, self.pseudo_inverse)

        output_data = UnmixedData(unmixed.astype(np.float32), self.chromophore_names,
                                  algorithm_id=self.algorithm_id,
                                  attributes=reconstruction.attributes,
                                  field_of_view=reconstruction.fov_3d)
        for a in reconstruction.attributes:
            output_data.attributes[a] = reconstruction.attributes[a]
        output_data.attributes[UnmixingAttributeTags.SUFFIX] = self.algorithm_id
        output_data.attributes[UnmixingAttributeTags.UNMIXING_WAVELENGTHS] = wavelengths
        output_data.attributes[UnmixingAttributeTags.SPECTRA] = self.chromophore_names
        output_data.hdf5_sub_name = reconstruction.hdf5_sub_name
        return output_data, {}, None

    def __init__(self, chromophores: List[Union[Spectrum, str]], wavelengths,
                 rescaling_factor=1, algorithm_id=""):
        super().__init__(algorithm_id)
        for i, c in enumerate(chromophores):
            if type(c) == str:
                chromophores[i] = SPECTRA_NAMES[c]
        spectra = np.array([c.get_spectrum(wavelengths) for c in chromophores])
        spectra = np.atleast_2d(spectra)
        self.chromophore_names = np.array([c.get_name() for c in chromophores])
        self.forward_matrix = spectra
        self.pseudo_inverse = np.linalg.pinv(spectra)
        self.wavelengths = np.array(wavelengths)
        self.rescaling_factor = rescaling_factor


class SO2Calculator(SpatialProcessingAlgorithm):
    """The SO2 calculator. This takes in unmixed data and produces SO2 data.
    """

    def __init__(self, algorithm_id="", nan_invalid=False):
        super().__init__(algorithm_id)
        self.nan_invalid = nan_invalid

    def run(self, spatial_data: UnmixedData, _, **kwargs):
        """
        Run the SO2 calculator.

        Parameters
        ----------
        spatial_data : UnmixedData
            The spatial data to process.
        _ : None
            Unused. This is here to make the interface consistent with the other algorithms.
        kwargs
            Unused.

        Returns
        -------
        tuple of (SingleImage, dict, None)
            The SO2 data, unused attributes, and unused by product. The first element is the only dataset that is used.
            The second two are there to make the interface consistent with the other algorithms.
        """
        hb_axis = np.where(spatial_data.spectra == "Hb")[0][0]
        hbo2_axis = np.where(spatial_data.spectra == "HbO2")[0][0]
        thb = spatial_data.raw_data[:, hb_axis] + spatial_data.raw_data[:, hbo2_axis]
        thb[thb == 0] = np.nan  # Just so it can pass tests.
        so2 = spatial_data.raw_data[:, hbo2_axis] / thb
        so2 = so2[:, None]
        if self.nan_invalid:
            so2[so2 > 1] = np.nan
            so2[so2 < 0] = np.nan
        output_data = SingleParameterData(so2.astype(np.float32), ["so2"],
                                          algorithm_id=self.algorithm_id,
                                          attributes=spatial_data.attributes,
                                          field_of_view=spatial_data.fov_3d)
        for a in spatial_data.attributes:
            output_data.attributes[a] = spatial_data.attributes[a]
        output_data.hdf5_sub_name = spatial_data.hdf5_sub_name
        return output_data, {}, None


class THbCalculator(SpatialProcessingAlgorithm):
    """The Total Haemoglobin calculator. This takes in unmixed data and produces THb data.
    """

    def run(self, spatial_data: UnmixedData, _, **kwargs):
        hb_axis = np.where(spatial_data.spectra == "Hb")[0][0]
        hbo2_axis = np.where(spatial_data.spectra == "HbO2")[0][0]
        thb = spatial_data.raw_data[:, hb_axis] + spatial_data.raw_data[:, hbo2_axis]
        thb = thb[:, None]
        output_data = SingleParameterData(thb.astype(np.float32), ["thb"],
                                          algorithm_id=self.algorithm_id,
                                          attributes=spatial_data.attributes,
                                          field_of_view=spatial_data.fov_3d)
        for a in spatial_data.attributes:
            output_data.attributes[a] = spatial_data.attributes[a]
        output_data.hdf5_sub_name = spatial_data.hdf5_sub_name
        return output_data, {}, None


class GasChallengeAnalyser(SpatialProcessingAlgorithm):
    """Analyser for oxygen-enhanced datasets. Takes in a time-series of sO2 data (or any other parameter) and produces
    a delta sO2.
    """

    def __init__(self, smoothing_window_size=10,
                 display_output=True, smoothing_sigma=2,
                 start_skip=0, challenge_type=1, buffer_width=5):
        """
        Gas challenge analyser.

        Parameters
        ----------
        smoothing_window_size : int
            The size of the window to use for smoothing the data.
        display_output : bool
            Whether to display the output, and ask for a response to confirm the fit is good.
        smoothing_sigma : float
            The sigma to use for the gaussian smoothing.
        start_skip : int
            The number of frames to skip at the start of the data.
        challenge_type : int
            The type of challenge to use. 1 is the standard challenge (air -> oxygen), -1 is the reverse (oxygen -> air).
        buffer_width : int
            The number of frames to use as a buffer at the start and end of the challenge.
        """
        super().__init__()
        self.smoothing_window_size = smoothing_window_size
        self.display = display_output
        self.smoothing_sigma = smoothing_sigma
        self.start_skip = start_skip
        self.sign = challenge_type
        self.buffer = buffer_width

    def run(self, so2: SingleParameterData, pa_data: PAData,
            reference_name=("reference_", "0"), **kwargs) -> Optional[Tuple[SingleImage, dict, Optional[List[
        ProcessingResult]]]]:
        """
        This algorithm takes in a so2 time series and a PA data object (pa_data) and returns a tuple containing the
        delta sO2 (change in sO2 between the baseline and the challenge), an empty dictionary (for compatibility with
        other processing methods), and a list containing the mean and standard deviation of the baseline sO2.

        Parameters
        ----------
        so2 : SingleParameterData
            The sO2 time series.
        pa_data : PAData
            The PA data object.
        reference_name : Tuple[str, str] or ROI
            The name of the reference region to use.
        kwargs : dict
            Additional keyword arguments, provided for compatibility with other processing methods.

        Returns
        -------
        Tuple[SingleImage, dict, Optional[List[ProcessingResult]]] or None
            A tuple containing the delta sO2, an empty dictionary (for compatibility with other processing methods), and
            a list containing the mean and standard deviation of the baseline sO2.
        """

        if type(reference_name) == ROI:
            roi_reference = reference_name
        else:
            rois = pa_data.get_rois()
            if not rois:
                raise RuntimeError("No regions of interest available.")
            elif reference_name not in rois:
                raise RuntimeError(f"{reference_name} not available.")
            else:
                roi_reference = rois[reference_name]

        roi_mask, _ = roi_reference.to_mask_slice(so2)
        steps = find_gc_boundaries(roi_mask, so2, self.smoothing_window_size,
                                   self.display, self.smoothing_sigma, self.start_skip,
                                   self.sign)
        if self.display:
            if input("Continue with analysis? ") not in ["Y", "y"]:
                if input("Set Manually? ") not in ["Y", "y"]:
                    return None
                else:
                    run = int(input("Enter Changeover Run Number: "))
                    steps[1] = run
                    steps = [steps[0], steps[1], steps[-1]]
                    print(steps)
        so2_measurements = so2.raw_data[:, 0]

        # SO2 in the baseline region
        baseline_region = so2_measurements[steps[0]: steps[1] - self.buffer]
        baseline_so2 = np.mean(baseline_region, axis=0)
        baseline_so2_std = np.std(baseline_region, axis=0)

        # SO2 in the challenge region
        post_so2 = so2_measurements[steps[1] + self.buffer: steps[2] - self.buffer]
        delta_so2 = (np.mean(post_so2, axis=0) - baseline_so2)

        delta_output = SingleImage(delta_so2, [HDF5Tags.DELTA_SO2],
                                   attributes=so2.attributes,
                                   field_of_view=so2.fov_3d)

        # Add meta data
        delta_output.hdf5_sub_name = so2.hdf5_sub_name
        delta_output.attributes[GCAttributeTags.STEPS] = steps
        delta_output.attributes[GCAttributeTags.BUFFER] = self.buffer
        delta_output.attributes[GCAttributeTags.SKIP_START] = self.start_skip

        # Add the extra outputs (baseline sO2 and baseline standard deviation)
        baseline_so2_output = SingleImage(baseline_so2, [HDF5Tags.BASELINE_SO2],
                                          attributes=delta_output.attributes,
                                          field_of_view=so2.fov_3d)
        baseline_so2_output.hdf5_sub_name = so2.hdf5_sub_name
        baseline_so2_sigma_output = SingleImage(baseline_so2_std, [HDF5Tags.BASELINE_SO2_STANDARD_DEVIATION],
                                                attributes=delta_output.attributes,
                                                field_of_view=so2.fov_3d)
        baseline_so2_sigma_output.hdf5_sub_name = so2.hdf5_sub_name
        return delta_output, {}, [baseline_so2_output, baseline_so2_sigma_output]


def find_dce_boundaries(roi_mask, icg, smoothing_window_size, display, smoothing_sigma):
    return find_gc_boundaries(roi_mask, icg, smoothing_window_size,
                              display, smoothing_sigma, 0, 2)


class DCEAnalyser(SpatialProcessingAlgorithm):
    """Analyser for DCE datasets. Takes in a time-series of ICG data and produces a delta ICG.
    """

    def __init__(self, smoothing_window_size=10,
                 display_output=True, smoothing_sigma=2,
                 buffer_width=5, unmix_index=2):
        super().__init__()
        self.smoothing_window_size = smoothing_window_size
        self.display = display_output
        self.smoothing_sigma = smoothing_sigma
        self.buffer = buffer_width
        self.unmix_index = unmix_index

    def run(self, unmixed_data: SingleParameterData, pa_data: PAData,
            reference_name=("reference_", "0"), **kwargs) -> Optional[Tuple[SingleImage, dict,
        Optional[List[
        ProcessingResult]]]]:
        if type(reference_name) == ROI:
            roi_reference = reference_name
        else:
            rois = pa_data.get_rois()
            if not rois:
                raise RuntimeError("No regions of interest available.")
            elif reference_name not in rois:
                raise RuntimeError(f"{reference_name} not available.")
            else:
                roi_reference = rois[reference_name]
        icg = unmixed_data[:, self.unmix_index]
        roi_mask, _ = roi_reference.to_mask_slice(icg)

        steps = find_dce_boundaries(roi_mask, icg, self.smoothing_window_size,
                                    self.display, self.smoothing_sigma)

        if self.display:
            if input("Continue with analysis? ") not in ["Y", "y"]:
                if input("Set Manually? ") not in ["Y", "y"]:
                    return None
                else:
                    run = int(input("Enter Changeover Run Number: "))
                    steps[1] = run
                    steps = [steps[0], steps[1], steps[-1]]
                    print(steps)

        icg_measurements = icg.raw_data
        baseline_region = icg_measurements[steps[0]: steps[1] - self.buffer]
        baseline_icg = np.mean(baseline_region, axis=0)
        baseline_icg_std = np.std(baseline_region, axis=0)
        post_icg = icg_measurements[steps[1]: steps[2] - self.buffer]
        delta_icg = np.max(post_icg, axis=0) - baseline_icg

        delta_output = SingleImage(delta_icg, [HDF5Tags.DELTA_ICG],
                                   attributes=icg.attributes,
                                   field_of_view=icg.fov_3d)
        delta_output.hdf5_sub_name = icg.hdf5_sub_name
        delta_output.attributes[GCAttributeTags.STEPS] = steps
        delta_output.attributes[GCAttributeTags.BUFFER] = self.buffer
        baseline_icg_output = SingleImage(baseline_icg, [HDF5Tags.BASELINE_ICG],
                                          attributes=delta_output.attributes,
                                          field_of_view=icg.fov_3d)
        baseline_icg_output.hdf5_sub_name = icg.hdf5_sub_name
        baseline_icg_sigma_output = SingleImage(baseline_icg_std, [HDF5Tags.BASELINE_ICG_SIGMA],
                                                attributes=delta_output.attributes,
                                                field_of_view=icg.fov_3d)
        baseline_icg_sigma_output.hdf5_sub_name = icg.hdf5_sub_name
        return delta_output, {}, [baseline_icg_output, baseline_icg_sigma_output]



#  Copyright (c) Thomas Else 2023.
#  License: MIT

from __future__ import annotations

import copy
from typing import Union, Dict, Tuple, Optional, Sequence, TYPE_CHECKING, Type

import h5py
import numpy as np
import xarray

from .hdf.fileimporter import ReaderInterface, WriterInterface
from .hdf.hdf5_interface import HDF5Reader, HDF5Writer
from ..core.image_structures.pa_time_data import PATimeSeries
from ..core.image_structures.single_image import SingleImage
from ..core.image_structures.single_parameter_data import SingleParameterData
from ..utils.roi_operations import get_rim_core_rois
from ..utils.rois.roi_type import ROI

if TYPE_CHECKING:
    try:
        from pyopencl.array import Array
    except ImportError:
        Array = np.ndarray
    from ..core.image_structures.image_sequence import ImageSequence

from functools import lru_cache

from ..io.attribute_tags import HDF5Tags


class PAData:
    """A class that contains the interface to access data from a single scan. Any source of scans (e.g.
    iThera/HDF5/IPASC) can be linked to this.
    """
    @property
    def shape(self) -> Tuple[int]:
        """
        Returns the shape of the dataset, minus the image size.

        Returns
        -------
        tuple of int
            Shape of the dataset.
        """
        return self.get_time_series().shape[:-2]

    def __getitem__(self, item: Union[slice, Tuple, None]) -> "PAData":
        """
        Slice the photoacoustic data. Choose a particular frame/wavelength etc.

        Parameters
        ----------
        item : slice or tuple

        Returns
        -------
        PAData
            Sliced pa data.
        """
        new_data = self.copy()
        new_data.scan_reader = copy.copy(new_data.scan_reader)[item]
        new_data.scan_writer = new_data.scan_writer
        return new_data

    def __init__(self, scan_reader: ReaderInterface, scan_writer: WriterInterface = None) -> None:
        """

        Parameters
        ----------
        scan_reader
        scan_writer
        """
        super().__init__()
        self.scan_reader = scan_reader
        self.scan_writer = scan_writer

        self.default_recon = None
        self.default_unmixing_type = ""
        self.external_roi_interface = None

    def is_clinical(self):
        return self.scan_reader.is_clinical()

    def set_default_recon(self, rec_name: Optional[str] = None) -> None:
        """
        Make all returned data be of a particular reconstruction type.
        It is recommended to run this at the start of analysis scripts.

        Parameters
        ----------
        rec_name : tuple of str, optional
        """
        if rec_name is None and self.default_recon is None:
            rec_name = list(self.get_scan_reconstructions().keys())[0]
        if not (self.default_recon is not None and rec_name is None):
            self.default_recon = rec_name

    def copy(self, cls: Optional[Type["PAData"]] = None) -> PAData:
        """
        Copy the pa data with changes given.

        Parameters
        ----------
        cls : Type[PAData]

        Returns
        -------
        PAData
            Copy of the dataset.
        """
        if cls is None:
            cls = self.__class__

        from copy import copy
        c = copy(self)
        c.__class__ = cls
        return c

    def get_scan_name(self) -> str:
        """
        Get the scan name.

        Returns
        -------
        str
            Scan name.

        """
        return self.scan_reader.get_scan_name()

    def get_scan_datetime(self):
        return self.scan_reader.get_scan_datetime()

    def get_sampling_frequency(self) -> float:
        """
        Get the scan's sampling frequency.

        Returns
        -------
        float
            Sampling Frequency
        """
        return self.scan_reader.get_sampling_frequency()

    def get_overall_correction_factor(self) -> Union[np.ndarray, Array, xarray.DataArray]:
        """
        Return the energy correction factors for the dataset.

        Returns
        -------
        np.ndarray
            Overall correction factor.
        """
        return self.scan_reader.get_correction_factor()

    def get_impulse_response(self) -> Union[np.ndarray, Array]:
        """
        Return the time-domain impulse response function.

        Returns
        -------
        np.ndarray or pyopencl.array.Array
            Impulse response function.
        """
        return self.scan_reader.get_impulse_response()

    def get_n_samples(self) -> int:
        """
        Get the number of time samples in the dataset.

        Returns
        -------
        int
            Number of samples.
        """
        return self.get_time_series().shape[-1]

    def get_speed_of_sound(self) -> Union[float, None]:
        """
        Get the speed of sound of the data if it has been set.

        Returns
        -------
        float or None
            Speed of sound
        """
        return self.scan_reader.get_speed_of_sound()

    def get_scan_geometry(self) -> Union[np.ndarray, Array]:
        """
        Get the scan detector geometry.

        Returns
        -------
        np.ndarray or pyopencl.array.Array
        """
        return self.scan_reader.get_sensor_geometry()

    def get_wavelengths(self) -> np.ndarray:
        """
        Get the wavelengths used in the scan.

        Returns
        -------
        np.ndarray
            Scan Wavelengths.
        """
        return self.scan_reader.get_wavelengths()

    def get_scan_images(self, group: str, ignore_default=False, suffix="") -> Union[
        Dict[Tuple[str, str], ImageSequence], ImageSequence]:
        """
        Get the scan images, e.g. reconstructions or so2 etc.

        Parameters
        ----------
        group : str
            Group to get images from.
        ignore_default : bool
            Ignore the default reconstruction.
        suffix : str
            Suffix to add to the image number (e.g. for ICG unmixing).

        Returns
        -------
        (dict of {tuple of (str, str): ImageSequence}) or ImageSequence
            Images of certain type if default recon has been set, or dict or images for all reconstructions.
        """

        datasets = self.scan_reader.get_datasets()
        if group not in datasets:
            return {}
        if self.default_recon is None or ignore_default:
            return datasets.get(group, {})
        else:
            image = (self.default_recon[0], self.default_recon[1] + suffix)
            return datasets.get(group, {}).get(image, {})

    def get_scan_reconstructions(self):
        return self.get_scan_images(HDF5Tags.RECONSTRUCTION)

    def get_scan_unmixed(self):
        return self.get_scan_images(HDF5Tags.UNMIXED, suffix=self.default_unmixing_type)

    def get_scan_so2(self):
        return self.get_scan_images(HDF5Tags.SO2)

    def close(self):
        self.scan_reader.close()
        if self.scan_writer is not None:
            self.scan_writer.close()

    def get_scan_mean(self, dataset: ImageSequence, operation=np.mean):
        """

        Parameters
        ----------
        dataset
        operation

        Returns
        -------

        """
        if type(dataset) == dict:
            raise NotImplementedError
        new_dataset = SingleImage(operation(dataset.raw_data, axis=0)[0], dataset.ax_1_labels,
                                  field_of_view=dataset.fov_3d, attributes=dataset.attributes)
        return new_dataset

    def get_scan_so2_time_mean(self):
        """

        Returns
        -------

        """
        return self.get_scan_mean(self.get_scan_so2())

    def get_scan_thb_time_mean(self):
        """

        Returns
        -------

        """
        return self.get_scan_mean(self.get_scan_thb())

    def get_scan_so2_time_standard_deviation(self):
        """

        Returns
        -------

        """
        return self.get_scan_mean(self.get_scan_so2(), np.std)

    def get_scan_thb(self):
        """

        Returns
        -------

        """
        return self.get_scan_images(HDF5Tags.THB)

    def get_scan_dso2(self):
        """

        Returns
        -------

        """
        return self.get_scan_images(HDF5Tags.DELTA_SO2)

    def get_scan_dicg(self):
        """

        Returns
        -------

        """
        return self.get_scan_images(HDF5Tags.DELTA_ICG)

    def get_scan_baseline_icg(self):
        """

        Returns
        -------

        """
        return self.get_scan_images(HDF5Tags.BASELINE_ICG)

    def get_scan_baseline_standard_deviation_icg(self):
        """

        Returns
        -------

        """
        return self.get_scan_images(HDF5Tags.BASELINE_ICG_SIGMA)

    def get_responding_pixels(self, nsigma=2):
        """

        Parameters
        ----------
        nsigma

        Returns
        -------

        """
        responding = None
        delta_images = None
        if self.get_scan_dicg():
            delta_images = self.get_scan_dicg()
            sigma_icg = self.get_scan_baseline_standard_deviation_icg()
            responding = delta_images.raw_data > nsigma * sigma_icg.raw_data
        elif self.get_scan_dso2():
            delta_images = self.get_scan_dso2()
            sigma_so2 = self.get_scan_baseline_standard_deviation_so2()
            responding = delta_images.raw_data > nsigma * sigma_so2.raw_data
        else:
            return None
        return SingleImage(responding, None, algorithm_id=delta_images.algorithm_id,
                           attributes=delta_images.attributes,
                           hdf5_sub_name=delta_images.hdf5_sub_name, field_of_view=delta_images.fov_3d)

    def get_scan_baseline_so2(self):
        """

        Returns
        -------

        """
        return self.get_scan_images(HDF5Tags.BASELINE_SO2)

    def get_scan_baseline_standard_deviation_so2(self):
        """

        Returns
        -------

        """
        return self.get_scan_images(HDF5Tags.BASELINE_SO2_STANDARD_DEVIATION)

    # Some cycling hypoxia analysis here.
    @lru_cache
    def get_scan_so2_frequency_components(self, do_detrend=True, fmin=1e-5, fmax=1000, fnum=1000):
        """

        Parameters
        ----------
        do_detrend
        fmin
        fmax
        fnum

        Returns
        -------

        """
        from scipy.signal import detrend, lombscargle
        so2 = self.get_scan_images(HDF5Tags.SO2)
        if type(so2) == dict:
            if len(so2) == 1:
                so2 = so2[list(so2.keys())[0]]
            else:
                raise NotImplementedError("""Frequency components are only enabled when there is one
                reconstruction set. Run PAData.set_default_recon() before running this if in doubt.
                                          """)

        detrended = detrend(so2.raw_data, axis=0, type="linear" if do_detrend else "constant")

        times = self.get_timestamps()[:, 0].copy()
        times -= times[0]

        frequencies = np.linspace(fmin, fmax, fnum)

        def lomb(a, b, c):
            return lombscargle(a, b.copy(), c)

        so2_frequency = SingleParameterData(np.apply_along_axis(lambda x: lomb(times, x, frequencies * 2 * np.pi),
                                                                0, detrended),
                                            so2.ax_1_labels,
                                            field_of_view=so2.fov_3d, attributes=so2.attributes)
        so2_frequency.da.coords["frames"] = frequencies * 2 * np.pi
        return so2_frequency

    def get_scan_so2_frequency_peak(self, fnum=1000):
        """

        Parameters
        ----------
        fnum

        Returns
        -------

        """
        so2 = self.get_scan_so2_frequency_components(fnum=fnum)
        raw_data = np.max(so2.raw_data, axis=0)[0]
        so2_frequency = SingleImage(raw_data,
                                            so2.ax_1_labels,
                                            field_of_view=so2.fov_3d, attributes=so2.attributes)
        return so2_frequency

    def get_scan_so2_frequency_sum(self, fnum=1000):
        """

        Parameters
        ----------
        fnum

        Returns
        -------

        """
        so2 = self.get_scan_so2_frequency_components(fnum=fnum)
        raw_data = np.sum(so2.raw_data, axis=0)[0]
        so2_frequency = SingleImage(raw_data,
                                            so2.ax_1_labels,
                                            field_of_view=so2.fov_3d, attributes=so2.attributes)
        return so2_frequency

    def get_segmentation(self):
        """

        Returns
        -------

        """
        return self.scan_reader.get_segmentation()

    def get_time_series(self) -> PATimeSeries:
        """

        Returns
        -------

        """
        dataset = self.scan_reader.get_pa_data()
        if type(dataset) is not PATimeSeries:
            raise ValueError("raw_data attribute must be either TimeSeries or FourierDomain type.")
        else:
            return dataset

    def get_recon_types(self) -> list:
        """
        Get the list of different reconstruction types.

        Returns
        -------
        list
            List of different reconstruction types that we have.
        """

        return list(self.get_scan_images(HDF5Tags.RECONSTRUCTION, True).keys())

    def get_z_positions(self) -> np.ndarray:
        """
        Get the z-positions of the sensor.

        Returns
        -------
        np.ndarray
            Z-positions array.
        """
        return self.scan_reader.get_scanner_z_position()

    def get_run_number(self) -> np.ndarray:
        """
        Get the run number for each of the frames.

        Returns
        -------
        np.ndarray
            Get the run numbers of each of the frames.

        """
        return self.scan_reader.get_run_numbers()

    def get_repetition_numbers(self) -> np.ndarray:
        """
        Get the scan repetition numbers for each frame.

        Returns
        -------
        np.ndarray
            Scan repetition numbers.
        """
        return self.scan_reader.get_repetition_numbers()

    def get_timestamps(self) -> np.ndarray:
        """
        Get the scan timestamps in seconds.

        Returns
        -------
        np.ndarray
            Timestamps in seconds.
        """
        # in seconds
        return self.scan_reader.get_scan_times()

    def get_rois(self, filter_rois=None,
                 interpolate: bool = False,
                 get_rim_cores=None,
                 rim_core_distance=None) -> Dict[Tuple[str, str], "ROI"]:
        """
        Get the regions of interest from the dataset.

        Parameters
        ----------
        rim_core_distance
        get_rim_cores
        filter_rois: dict or None
        interpolate: bool

        Returns
        -------
        dict of {(tuple of (str, str): ROI}
            Return all the rois.

        """
        if self.external_roi_interface is not None:
            reader = self.external_roi_interface.scan_reader
        else:
            reader = self.scan_reader
        output_rois = reader.get_rois(interpolate)

        if get_rim_cores is not None:
            rim_core_rois = {}
            for o in output_rois:
                if o[0] in get_rim_cores:
                    core, rim = get_rim_core_rois(output_rois[o], rim_core_distance)
                    name, position = o[0].split("_")
                    rim_core_rois[(name + ".core_" + position, o[1])] = core
                    rim_core_rois[(name + ".rim_" + position, o[1])] = rim
            output_rois.update(rim_core_rois)
        if filter_rois is not None:
            new_out_rois = {}
            for k in output_rois:
                for f in filter_rois:
                    if f == k[0] or f + "_" == k[0]:
                        new_out_rois[k] = output_rois[k]
            output_rois = new_out_rois
        return output_rois

    def delete_recons(self, name=None, recon_groups: Optional[Sequence[str]] = None):
        """
        Delete the reconstructions.

        Parameters
        ----------
        name : str or None
        recon_groups : (iterable of str) or None
        """
        if self.scan_writer is None:
            raise NotImplementedError("Deletion only possible with a writing interface.")
        self.scan_writer.delete_recons(name, recon_groups)

    def set_speed_of_sound(self, c: float) -> None:
        """
        Change the speed of sound for the dataset.

        Parameters
        ----------
        c : float
            Speed of sound.

        """
        if self.scan_writer is None:
            raise NotImplementedError("No writing capability enabled.")
        self.scan_writer.set_speed_of_sound(c)

    def delete_rois(self, name_position: Optional[str] = None, number: Optional[str] = None) -> None:
        """
        Delete a roi with name and number. If number is None, will delete all.
        If name and number is None, delete all.

        Parameters
        ----------
        name_position
            str or None
        number
            str or none
        """
        self.scan_writer.delete_rois(name_position, number)

    def add_roi(self, roi_data: "ROI", generated: bool = False) -> None:
        """
        Add a region of interest to the hdf5 file.

        Parameters
        ----------
        roi_data : ROI
        generated : bool, default False
        """
        self.scan_writer.add_roi(roi_data, generated)

    def rename_roi(self, old_name: Union[str, Tuple], new_name: str, new_position: str) -> None:
        """
        Rename a region of interest.

        Parameters
        ----------
        old_name : str or tuple
            Old roi name e.g. "tumour_left/0" or ("tumour_left", "0")
        new_name : str
            New roi name e.g. "brain"
        new_position : str
            New roi position e.g. "left"

        """
        self.scan_writer.rename_roi(old_name, new_name, new_position)

    def clear_dso2(self) -> None:
        self.scan_writer.delete_dso2s()

    @classmethod
    def from_hdf5(cls, filename: Union[str, h5py.File], mode: str = "r"):
        try:
            file = h5py.File(filename, mode)
        except TypeError:
            file = filename
        return cls(HDF5Reader(file), HDF5Writer(file))

    def save_hdf5(self, filename: str):
        file = h5py.File(filename, "a")
        writer = HDF5Writer(file)
        return writer.save_file(self.scan_reader)

    def save_to_hdf5(self, filename):
        return self.save_hdf5(filename)

    @property
    def dataset(self):
        """

        Returns
        -------

        """
        return self.get_time_series()

    def summary_measurements(self, metrics=None,
                             include_rois=None, roi_kwargs=None, just_summary=True, return_masks=False):
        """

        Parameters
        ----------
        metrics
        include_rois
        roi_kwargs
        just_summary

        Returns
        -------

        """
        import pandas as pd
        if metrics is None:
            metrics = ["thb", "so2"]
        if roi_kwargs is None:
            roi_kwargs = {}

        rois = self.get_rois(**roi_kwargs)

        if include_rois is not None:
            new_rois = {}
            for x in rois:
                if x[0][:-1] in include_rois or x[0] in include_rois:
                    new_rois[x] = rois[x]
            rois = new_rois

        measurements = []
        for m in metrics:
            if m == "thb":
                measurements.append(self.get_scan_thb())
            elif m == "so2":
                measurements.append(self.get_scan_so2())
            elif m == "icg":
                unmixed = self.get_scan_unmixed()
                n = [i for i, l in enumerate(unmixed.ax_1_labels) if l.lower() == "icg"]
                if len(n) == 0:
                    print("No icg channel found")
                    continue
                n= n[0]
                measurements.append(unmixed[:, n: n + 1])
            elif m == "dso2":
                measurements.append(self.get_scan_dso2())
            elif m == "baseline_so2":
                measurements.append(self.get_scan_baseline_so2())
            elif m == "responding":
                measurements.append(self.get_responding_pixels())
            elif m == "recons":
                measurements.append(self.get_scan_reconstructions())
            else:
                raise NotImplementedError(f"Metric {m} not yet implemented.")

        if not rois:
            return pd.DataFrame({})
        outputs = []
        for name, roi in rois.items():
            output_roi = {}
            for i, m in enumerate(measurements):
                if m:
                    mask, data_slice = roi.to_mask_slice(m)
                    output_roi[metrics[i]] = data_slice.raw_data.T[mask.T].T
                else:
                    print(f"Skipping metric {metrics[i]}")

            mask, _, selection = roi.to_mask_slice(self.get_scan_reconstructions(), return_selection=True)
            if return_masks:
                return_mask, _ = roi.to_mask_slice(self.get_scan_so2())
                output_roi["Mask"] = return_mask
            output_roi["Timings"] = self.get_timestamps()[selection]
            output_roi["Area"] = np.sum(mask)
            for a in roi.attributes:
                output_roi[a] = roi.attributes[a]
            output_roi["number"] = name[1]
            output_roi["name"] = name
            output_roi["Wavelengths"] = self.get_wavelengths()
            outputs.append(output_roi)

        output_table = pd.DataFrame(outputs)

        summary_methods = {"mean": np.mean,
                           "median": np.median,
                           "std": np.std}
        for name, method in summary_methods.items():
            for metric in metrics:
                if metric in output_table.columns:
                    output_table[metric + "_" + name] = output_table[metric].apply(
                        lambda t: np.squeeze(method(t, axis=-1))[()])

        if just_summary:
            for metric in metrics:
                if metric in output_table.columns:
                    del output_table[metric]

        return output_table



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import numpy as np
from scipy.signal import fftconvolve


def find_gc_boundaries(mask, data_so2,
                       window=10, display=False, sigma=2, skip_start=0,
                       sign=1):
    """
    Find the runs in the time series at which the breathing gas of the animal was changed.

    Parameters
    ----------
    mask : array_like
        A boolean array which can be applied to the time series.
    data_so2 : SingleParameterData
        The time series of the SO2 data.
    window : int, optional
        The window size to use for the convolution.
    display : bool, optional
        Whether to display the results.
    sigma : float, optional
        The sigma to use for the convolution.
    skip_start : int, optional
        The number of runs to skip at the start of the time series.
    sign : int, optional
        The sign of the change in the time series.

    Returns
    -------
    list
        List of integers with the indices of the detected gas changes.
    """

    # STEP 1: Smooth the curve with a Gaussian kernel
    # Complete error here:
    measurement = np.squeeze(data_so2.raw_data.T[mask.T].T)

    if sign != 2:
        measurement *= sign

    kernel = np.arange(window) - window / 2 + 1 / 2
    kernel = np.exp(-(kernel / sigma) ** 2)
    kernel /= np.sum(kernel)
    kernel = kernel[:, None]

    from scipy.ndimage import median_filter
    smoothed = median_filter(measurement, kernel.shape)
    smoothed = fftconvolve(smoothed, kernel, "valid")

    # Find the gradient of the curve (to give us the maximum).
    smoothed_grad = np.median(np.gradient(smoothed, axis=0), axis=-1)
    steps = [skip_start]
    if sign == 2:
        smoothed_grad = smoothed_grad ** 2
    # Find first peak in derivative.
    step_point = np.argmax(smoothed_grad)
    steps.append(step_point + window // 2)
    step_grad = smoothed_grad[step_point]

    # Find second peak in the derivative
    step_point_b = step_point + np.argmin(smoothed_grad[step_point:])
    step_grad_b = smoothed_grad[step_point_b]

    # Custom stuff to try and avoid false positives (this could be improved somehow).
    if step_grad_b < - step_grad / 2 and step_point_b - step_point > step_point * 0.5:
        # I.e. detect if there actually was a second change.
        steps.append(step_point_b + window // 2)
        steps.append(len(measurement) - 1)
    else:
        steps.append(len(measurement) - 1)

    # Display the changeover points.
    if display:
        import matplotlib.pyplot as plt
        plt.plot(np.median(smoothed * sign, axis=-1), label="Gas Challenge Trace in the Reference Region")
        plt.twinx()
        plt.plot(smoothed_grad * sign, c="C1", label="Derivative of Gas Challenge Trace in the Reference Region")
        for s in steps:
            plt.axvline(s - window // 2)
        plt.gcf().legend()
        plt.show()
    return steps



#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
``patato-reconstruct``: A script to reconstruct all MSOT data files in a folder.
"""

import argparse
import glob
import json
import logging
import os

from .. import PAData
from ..recon import get_default_recon_preset
from .. import read_reconstruction_preset
from ..utils import sort_key
from ..utils.pipeline import run_pipeline


def main():
    parser = argparse.ArgumentParser(description="Process MSOT Data.")
    parser.add_argument('input', type=str, help="Data File")
    parser.add_argument('-c', '--speed', type=float, help="Speed of Sound", default=None)
    parser.add_argument('-o', '--output', type=str, help="Output File", default=None)
    parser.add_argument('-p', '--preset', type=str, help="Preset File",
                        default=None)
    parser.add_argument('-r', '--run', type=int, help="Run Number",
                        default=None)
    parser.add_argument('-w', '--wavelength', type=int, help="Wavelength Number",
                        default=None)
    parser.add_argument('-R', '--repeat', type=lambda x: x in ["True", "true", "T", "Y", "y", "yes", "YES", "1"],
                        default=True, help="Repeat recon if already exists.")
    parser.add_argument('-H', '--highpass', type=float, help="High pass filter override", default=None)
    parser.add_argument('-L', '--lowpass', type=float, help="Low pass filter override", default=None)
    parser.add_argument('-C', '--clear', type=bool, default=False, help="Clear all existing recons.")
    parser.add_argument('-I', '--ipasc', type=bool, default=False, help="Export to IPASC format as well.")
    parser.add_argument('-d', '--debug', type=bool, default=False, help="Enable debugging logging.")
    parser.add_argument(
        "-v", "--version", action="version",
        version=f"{parser.prog} version 0.1"
    )
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    files = []

    data_file = args.input
    if os.path.isfile(data_file):
        files.append(data_file)
    elif os.path.isdir(data_file):
        files = sorted(glob.glob(os.path.join(data_file, "**", "*.hdf5"), recursive=True), key=sort_key)

    output_file = args.output

    clear_recons = args.clear
    recompute_recon = args.repeat
    run_number = args.run
    if run_number is None:
        run_number = slice(None, None)
    else:
        run_number = slice(run_number, run_number + 1)
    wavelength = args.wavelength
    if wavelength is None:
        wavelength = slice(None, None)
    else:
        wavelength = slice(wavelength, wavelength + 1)
    do_ipasc_export = args.ipasc

    print(f"Identified {len(files)} files to process!")

    for data_file in files:
        # Run processing.
        pa_data = PAData.from_hdf5(data_file, "r+")[run_number, wavelength]

        json_path = args.preset
        if json_path is None:
            json_path = get_default_recon_preset(pa_data)

        with open(json_path) as json_file:
            settings = json.load(json_file)

        if args.lowpass is not None:
            settings["FILTER_LOW_PASS"] = args.lowpass
        elif args.highpass is not None:
            settings["FILTER_HIGH_PASS"] = args.highpass

        # Setup the reconstruction pipeline
        pipeline = read_reconstruction_preset(settings)

        # Set the reconstruction description
        description = pipeline.children[0].get_algorithm_name()

        # Setup the name
        if args.highpass is not None:
            description += "_CUSTOM_HP_FILTER"
        if args.lowpass is not None:
            description += "_CUSTOM_LP_FILTER"
        if args.run is not None:
            description += "_run_" + str(args.run)
        if args.wavelength is not None:
            description += "_wavelength_" + str(args.run)

        # Start processing
        if clear_recons:
            pa_data.delete_recons()

        if description in pa_data.get_recon_types() and not recompute_recon:
            print("Already computed recon", description, "for", data_file + ".")
            continue

        print(data_file, pa_data.dataset.shape)
        if any(x == 0 for x in pa_data.dataset.shape):
            continue

        run_pipeline(pipeline, pa_data.dataset, pa_data, n_batch=1 + 20 // pa_data.shape[1], output_file=output_file)

        if do_ipasc_export:
            from .. import export_to_ipasc
            export_to_ipasc(data_file)


if __name__ == "__main__":
    main()



#  Copyright (c) Thomas Else 2023.
#  License: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Sequence

import h5py
import numpy as np
from ...core.image_structures.image_structure_types import IMAGE_DATA_TYPES
from ...core.image_structures.pa_time_data import PATimeSeries
from ...io.attribute_tags import HDF5Tags, ROITags, UnmixingAttributeTags, ReconAttributeTags
from ...io.hdf.fileimporter import WriterInterface, ReaderInterface
from ...utils.rois.roi_type import ROI
import json

if TYPE_CHECKING:
    import numpy as np
import dask.array as da


def renumber_group(dataset: h5py.Group) -> None:
    """
    Renumber an old roi group.

    Parameters
    ----------
    dataset : h5py.Group
    """
    # Clean up old group:
    original_names = sorted(list(dataset.keys()), key=int)
    new_names = [str(x) for x in range(len(original_names))]
    for old, new in zip(original_names, new_names):
        dataset.move(old, new)


# Might be useful to set enviro variable: HDF5_USE_FILE_LOCKING=FALSE
def load_image_from_hdf5(cls, dataset, file):
    """

    Parameters
    ----------
    cls
    dataset
    file

    Returns
    -------

    """
    ax_1_meaning = cls.get_ax1_label_meaning()
    dataset_da = da.from_array(dataset, chunks=(1,) + dataset.shape[1:])

    if ax_1_meaning == HDF5Tags.WAVELENGTH:
        ax_1_labels = dataset.attrs.get(HDF5Tags.WAVELENGTH, file["wavelengths"])
    elif ax_1_meaning == "SPECTRA":
        ax_1_labels = dataset.attrs["SPECTRA"]
    elif ax_1_meaning is None:
        # For dso2 etc.
        ax_1_labels = np.array([dataset.name.split("/")[-3]])
    elif dataset_da.shape[1] == 1:
        ax_1_labels = np.array([dataset.name.split("/")[-3]])
    else:
        ax_1_labels = np.arange(dataset_da.shape[1])

    fov = [dataset.attrs.get(fx, None) for fx in [ReconAttributeTags.X_FIELD_OF_VIEW,
                                                  ReconAttributeTags.Y_FIELD_OF_VIEW,
                                                  ReconAttributeTags.Z_FIELD_OF_VIEW]]

    if all([x is None for x in fov]):
        fov_x = dataset.attrs.get(ReconAttributeTags.OLD_FIELD_OF_VIEW, 1.)
        fov = [fov_x, fov_x, fov_x]

    new_cls = cls(dataset_da, ax_1_labels,
                  algorithm_id=dataset.attrs.get(UnmixingAttributeTags.SUFFIX, ""),
                  attributes=dict(dataset.attrs),
                  field_of_view=fov)
    if not cls.is_single_instance():
        new_cls.hdf5_sub_name = dataset.name.split("/")[-2]
    return new_cls


class HDF5Writer(WriterInterface):
    def delete_images(self, image):
        pass

    def __init__(self, file):
        WriterInterface.__init__(self)
        if type(file) == h5py.File:
            self.file = file
        else:
            self.file = h5py.File(file, "a")
        self.overwrite = False  # TODO: implement this.

    def set_scan_datetime(self, datetime):
        self.file.attrs[HDF5Tags.DATE] = str(datetime)

    def set_pa_data(self, raw_data):
        if type(raw_data) == PATimeSeries:
            raw_data = raw_data.raw_data
        if raw_data is None:
            return
        self.file.create_dataset(HDF5Tags.RAW_DATA, data=raw_data)

    def set_scan_name(self, scan_name: str):
        self.file.attrs[HDF5Tags.SCAN_NAME] = scan_name
        self.file[HDF5Tags.RAW_DATA].attrs[HDF5Tags.SCAN_NAME] = scan_name

    def set_temperature(self, temperature: "np.ndarray"):
        self.file.create_dataset(HDF5Tags.TEMPERATURE, data=temperature, compression="gzip")

    def set_correction_factor(self, correction_factor):
        self.file.create_dataset(HDF5Tags.OVERALL_CORR, data=correction_factor, compression="gzip")

    def set_scanner_z_position(self, z_position):
        self.file.create_dataset(HDF5Tags.Z_POSITION, data=z_position, compression="gzip")

    def set_run_numbers(self, run_numbers):
        self.file.create_dataset(HDF5Tags.RUN, data=run_numbers, compression="gzip")

    def set_repetition_numbers(self, repetition_numbers):
        self.file.create_dataset(HDF5Tags.REPETITION, data=repetition_numbers, compression="gzip")

    def set_scan_times(self, scan_times):
        self.file.create_dataset(HDF5Tags.TIMESTAMP, data=scan_times, compression="gzip")

    def set_sensor_geometry(self, sensor_geometry):
        self.file.create_dataset(HDF5Tags.SCAN_GEOMETRY, data=sensor_geometry, compression="gzip")

    def set_us_data(self, us_data, us_fov):
        # At some point convert this to image sequence data types.
        self.file.create_dataset(HDF5Tags.ULTRASOUND, data=us_data)
        self.file[HDF5Tags.ULTRASOUND].attrs[HDF5Tags.ULTRASOUND_FIELD_OF_VIEW] = us_fov

    def set_impulse_response(self, impulse_response):
        self.file.create_dataset(HDF5Tags.IMPULSE_RESPONSE, data=impulse_response, compression="gzip")

    def set_wavelengths(self, wavelengths):
        self.file.create_dataset(HDF5Tags.WAVELENGTH, data=wavelengths)

    def set_us_offsets(self, us_offsets):
        if us_offsets is not None:
            self.file.create_dataset(HDF5Tags.ULTRASOUND_FRAME_OFFSET, data=us_offsets,
                                     compression="gzip")

    def set_water_absorption(self, water_absorption, pathlength):
        # TODO: Maybe refactor this?
        water = self.file.create_dataset(HDF5Tags.WATER_ABSORPTION_COEFF, data=water_absorption,
                                         compression="gzip")
        water.attrs[HDF5Tags.WATER_PATHLENGTH] = pathlength

    def add_image(self, image_data):
        # Image data can be reconstructions, unmixed data etc.
        grp = self.file.require_group(image_data.get_hdf5_group_name())  # E.g. unmixed or recons
        subgroup = grp.require_group(image_data.get_hdf5_sub_name())  # E.g. reconstruction method or something
        name = str(len(subgroup.keys())) + image_data.algorithm_id  # E.g. "0" + "E"
        dataset = subgroup.create_dataset(name, data=image_data.raw_data)

        # Copy attributes
        for a in image_data.attributes:
            if type(image_data.attributes[a]) == dict:
                for b in image_data.attributes[a]:
                    attr = image_data.attributes[a][b]
                    if type(attr) == str:
                        attr = bytes(attr, "utf-8")
                    try:
                        dataset.attrs[b] = attr
                    except TypeError:
                        dataset.attrs[b] = json.dumps(attr)
            else:
                attr = image_data.attributes[a]
                if type(attr) == str:
                    attr = bytes(attr, "utf-8")
                from numpy import nan, ndarray
                if attr is None:
                    attr = nan
                if type(attr) is ndarray:
                    if attr.dtype == "<U4":
                        attr = [bytes(a, "utf-8") for a in attr]
                dataset.attrs[a] = attr

    def set_scan_comment(self, comment: str):
        self.file.attrs[HDF5Tags.SCAN_COMMENT] = comment

    def set_sampling_frequency(self, frequency: float):
        self.file[HDF5Tags.RAW_DATA].attrs[HDF5Tags.SAMPLING_FREQ] = frequency
        self.file.attrs[HDF5Tags.SAMPLING_FREQ] = frequency

    def set_segmentation(self, seg):
        self.file.create_dataset(HDF5Tags.SEGMENTATION, data=seg, compression="gzip")

    def set_speed_of_sound(self, c: float):
        self.file[HDF5Tags.RAW_DATA].attrs[HDF5Tags.SPEED_OF_SOUND] = c
        self.file.attrs[HDF5Tags.SPEED_OF_SOUND] = c

    def add_roi(self, roi_data: ROI, generated: bool = False):
        """
        Add a region of interest to the hdf5 file.

        Parameters
        ----------
        roi_data : ROI
        generated : bool, default False
        """
        roi_group = self.file.require_group(HDF5Tags.REGIONS_OF_INTEREST)
        print(roi_data.attributes[ROITags.ROI_NAME] + "_" +
              roi_data.attributes[ROITags.ROI_POSITION])
        region_group = roi_group.require_group(roi_data.attributes[ROITags.ROI_NAME] + "_" +
                                               roi_data.attributes[ROITags.ROI_POSITION])
        n = str(len(region_group))
        if type(roi_data.points) in [np.ndarray, list]:
            dataset = region_group.create_dataset(n, data=roi_data.points)
        else:
            raise NotImplementedError("Cannot save specialist roi type (yet...).")
        # Add attributes
        for attr in roi_data.attributes:
            dataset.attrs[attr] = roi_data.attributes[attr]
        dataset.attrs[ROITags.GENERATED_ROI] = generated

    def rename_roi(self, old_name, new_name, new_position):
        if type(old_name) == str:
            old_name = tuple(old_name.split("/"))
        roi_dataset = self.file[HDF5Tags.REGIONS_OF_INTEREST]
        roi_dataset["/".join(old_name)].attrs[ROITags.ROI_NAME] = new_name
        roi_dataset["/".join(old_name)].attrs[ROITags.ROI_POSITION] = new_position

        new_dataset_name = new_name + "_" + new_position
        new_number = "0"

        if new_dataset_name in roi_dataset:
            new_number = str(len(roi_dataset[new_dataset_name]))
        roi_dataset.move("/".join(old_name), new_dataset_name + "/" + new_number)

        renumber_group(roi_dataset[old_name[0]])

    def delete_rois(self, name_position=None, number=None):
        if "/" in name_position and number is None:
            name, number = name_position.split("/")
            name_position = name

        if HDF5Tags.REGIONS_OF_INTEREST in self.file:
            if name_position is None and number is None:
                del self.file[HDF5Tags.REGIONS_OF_INTEREST]
            elif number is None and name_position in self.file[HDF5Tags.REGIONS_OF_INTEREST]:
                del self.file[HDF5Tags.REGIONS_OF_INTEREST][name_position]
            elif name_position in self.file[HDF5Tags.REGIONS_OF_INTEREST]:
                if number in self.file[HDF5Tags.REGIONS_OF_INTEREST][name_position]:
                    del self.file[HDF5Tags.REGIONS_OF_INTEREST][name_position][number]
                    # Renumber rois.
                    renumber_group(self.file[HDF5Tags.REGIONS_OF_INTEREST][name_position])

    def delete_recons(self, name=None, recon_groups: Optional[Sequence[str]] = None):
        """
        Delete some reconstructions.

        Parameters
        ----------
        name : str or None
        recon_groups : (iterable of str) or None
        """
        if recon_groups is None:
            recon_groups = [HDF5Tags.RECONSTRUCTION,
                            HDF5Tags.UNMIXED,
                            HDF5Tags.SO2,
                            HDF5Tags.THB,
                            HDF5Tags.DELTA_SO2,
                            HDF5Tags.DELTA_ICG,
                            HDF5Tags.BASELINE_ICG,
                            HDF5Tags.BASELINE_SO2,
                            HDF5Tags.BASELINE_SO2_STANDARD_DEVIATION,
                            HDF5Tags.BASELINE_ICG_SIGMA]
        if name is None:
            for group in recon_groups:
                if group in self.file:
                    del self.file[group]
        else:
            for group in recon_groups:
                if group in self.file:
                    for recon_name in self.file[group]:
                        if name == recon_name:
                            del self.file[group][name]

    def delete_dso2s(self):
        """
        Delete the old dso2 datasets.
        Might want to remove this or refactor it in some way?
        """
        for group in [HDF5Tags.DELTA_SO2, HDF5Tags.BASELINE_SO2_STANDARD_DEVIATION, HDF5Tags.BASELINE_SO2,
                      HDF5Tags.DELTA_ICG, HDF5Tags.BASELINE_ICG, HDF5Tags.BASELINE_ICG_SIGMA]:
            if group in self.file:
                del self.file[group]


class HDF5Reader(ReaderInterface):
    def _get_rois(self):
        output = {}
        if HDF5Tags.REGIONS_OF_INTEREST in self.file:
            for roi_name in self.file[HDF5Tags.REGIONS_OF_INTEREST]:
                roi_group = self.file[HDF5Tags.REGIONS_OF_INTEREST][roi_name]
                for roi_number in roi_group:
                    dataset = roi_group[roi_number]
                    clinical = self.is_clinical()
                    frame_type = ROITags.Z_POSITION if not clinical else ROITags.REPETITION
                    match_frames = self.get_scanner_z_position() if not clinical else self.get_repetition_numbers()
                    ax0_indices = np.where(np.isclose(match_frames[:, 0], dataset.attrs.get(frame_type, 1.0)))[0]
                    output[(roi_name, roi_number)] = ROI(dataset[:], dataset.attrs[ROITags.Z_POSITION],
                                                         dataset.attrs[ROITags.RUN],
                                                         dataset.attrs.get(ROITags.REPETITION, np.nan),
                                                         dataset.attrs[ROITags.ROI_NAME],
                                                         dataset.attrs[ROITags.ROI_POSITION],
                                                         dataset.attrs.get(ROITags.GENERATED_ROI, False),
                                                         ax0_indices)
        return output

    def __init__(self, file):
        ReaderInterface.__init__(self)
        if type(file) == h5py.File:
            self.file = file
        else:
            self.file = h5py.File(file, "r")

    def get_scan_datetime(self):
        import dateutil.parser
        try:
            if type(self.file.attrs[HDF5Tags.DATE]) == str:
                return dateutil.parser.isoparse(self.file.attrs[HDF5Tags.DATE]).replace(tzinfo=None)
            else:
                return self.file.attrs[HDF5Tags.DATE]
        except KeyError:
            return np.nan

    def _get_pa_data(self):
        return self.file[HDF5Tags.RAW_DATA], dict(self.file[HDF5Tags.RAW_DATA].attrs)

    def get_scan_name(self):
        return self.file[HDF5Tags.RAW_DATA].attrs[HDF5Tags.SCAN_NAME]

    def _get_temperature(self):
        return self.file[HDF5Tags.TEMPERATURE]

    def _get_correction_factor(self):
        if np.any(np.isnan(self.file[HDF5Tags.OVERALL_CORR][:])):
            # Old version
            return self.file["POWER"][:]
        return self.file[HDF5Tags.OVERALL_CORR][:]

    def _get_scanner_z_position(self):
        return self.file[HDF5Tags.Z_POSITION]

    def _get_run_numbers(self):
        return self.file[HDF5Tags.RUN]

    def _get_repetition_numbers(self):
        return self.file[HDF5Tags.REPETITION]

    def _get_scan_times(self):
        return self.file[HDF5Tags.TIMESTAMP][:, :] * 1e-7

    def _get_sensor_geometry(self):
        return self.file[HDF5Tags.SCAN_GEOMETRY][:]

    def get_us_data(self):
        if HDF5Tags.ULTRASOUND in self.file:
            return self.file.get(HDF5Tags.ULTRASOUND, None), \
                   self.file[HDF5Tags.ULTRASOUND].attrs[HDF5Tags.ULTRASOUND_FIELD_OF_VIEW]
        else:
            return None, {}

    def get_us_offsets(self):
        return self.file[HDF5Tags.ULTRASOUND_FRAME_OFFSET]

    def get_impulse_response(self):
        return self.file[HDF5Tags.IMPULSE_RESPONSE]

    def _get_wavelengths(self):
        return self.file[HDF5Tags.WAVELENGTH][:]

    def _get_water_absorption(self):
        return (self.file[HDF5Tags.WATER_ABSORPTION_COEFF],
                self.file[HDF5Tags.WATER_ABSORPTION_COEFF].attrs[HDF5Tags.WATER_PATHLENGTH])

    def _get_datasets(self):
        output = {}
        for image_type in IMAGE_DATA_TYPES:
            output[image_type] = {}
            dtype = IMAGE_DATA_TYPES[image_type]
            images = []
            if image_type in self.file:
                # Loop through all types of this image and add them to the list
                for recon in self.file[image_type]:
                    for recon_num in self.file[image_type][recon]:
                        image = load_image_from_hdf5(dtype, self.file[image_type][recon][recon_num], self.file)
                        images.append(((recon, recon_num), image))
            for (recon, num), data in images:
                output[image_type][(recon, num)] = data
        return output

    def get_scan_comment(self):
        return self.file.attrs[HDF5Tags.SCAN_COMMENT]

    def _get_sampling_frequency(self):
        return self.file[HDF5Tags.RAW_DATA].attrs[HDF5Tags.SAMPLING_FREQ]

    def get_speed_of_sound(self):
        return self.file[HDF5Tags.RAW_DATA].attrs.get(HDF5Tags.SPEED_OF_SOUND, None)

    def _get_segmentation(self):
        return self.file.get(HDF5Tags.SEGMENTATION, None)


class PACFISHInterface(ReaderInterface):
    pass



try:
    import cupy as cp
    from cupyx.scipy.sparse import csr_matrix, vstack
    cupy_enabled = True
except ImportError:
    cupy_enabled = False

#  Copyright (c) Thomas Else 2023.
#  License: MIT

from os.path import dirname, join, exists

import numpy as np


def generate_model(det_x, det_y, dl, dx, nx, x_0, nt):
    # TODO: validate types.
    # Load the cuda code:
    directory = dirname(__file__)

    with open(join(directory, "generate_model.cu"), "r") as w:
        cuda_code = w.read()
    calculate_element = cp.RawKernel(cuda_code, 'calculate_element', jitify=True)

    # TODO: allow non-square reconstruction areas.
    ntpixel = cp.int32(4 * np.sqrt(2) * dx / dl)

    # Normalise the values. Convert to appropriate format.
    det_x /= dl
    det_x = cp.double(det_x)
    det_y /= dl
    det_y = cp.double(det_y)
    x_0 /= dl
    x_0 = cp.double(x_0)
    dx /= dl
    dx = cp.double(dx)

    matrices = []
    i = 0

    output = cp.zeros(int(ntpixel * nx * nx), dtype=cp.float64)
    indices = cp.zeros(int(ntpixel * nx * nx), dtype=cp.int32)

    positions = cp.repeat(cp.arange(nx * nx)[:, None], ntpixel, axis=-1).flatten()

    dl = cp.float64(1.)
    for detx, dety in zip(det_x, det_y):
        i += 1
        # TODO: optimise the block/grid size below.
        calculate_element((128,), (128,), (output, indices, nx,
                                           ntpixel,
                                           detx, dety, dl, x_0, dx))
        matrix = csr_matrix((output.flatten(), (indices.flatten(), positions)), shape=(nt, nx * nx))
        matrices.append(matrix)
    m = vstack(matrices)
    return m


def get_hash(*x):
    to_hash = []
    for y in x:
        if type(y) == np.ndarray:
            y = tuple(y.flatten())
        to_hash.append(y)
    h = hash(tuple(to_hash))
    return hex(np.uint64(h))


def get_model(det_x, det_y, dl, dx, nx, x_0, nt,
              cache=True, hash_fn=None):
    det_x = det_x.astype(np.float64)
    det_y = det_y.astype(np.float64)
    dl = cp.float64(dl)
    dx = cp.float64(dx)
    nx = cp.int32(nx)
    x_0 = cp.float64(x_0)
    nt = cp.int32(nt)
    print("Loading model")
    if hash_fn is None:
        hash_fn = get_hash
    if cache:
        h = hash_fn(det_x, det_y, dl, dx, nx, x_0, nt)
        model_folder = join(dirname(__file__), "models")
        filename = join(model_folder, h + ".npz")
        import scipy.sparse
        if exists(filename):
            mat = csr_matrix(scipy.sparse.load_npz(filename))
        else:
            mat = generate_model(det_x, det_y, dl, dx, nx, x_0, nt)
            scipy.sparse.save_npz(filename, mat.astype(cp.float32).get(), compressed=False)
        return mat
    else:
        return generate_model(det_x, det_y, dl, dx, nx, x_0, nt)


def test_forward_model(hash_fn=None):
    import matplotlib.pyplot as plt
    dl = cp.float64(1540 / 4e7)

    det_thetas = np.linspace(cp.pi / 4, 7 * cp.pi / 4, 256)
    detectors = np.array([np.cos(det_thetas), np.sin(det_thetas)]).T * 0.0405

    lx = cp.float64(0.025)
    x_0 = -lx / 2
    nx = 512
    dx = lx / (nx - 1)
    nt = 2030

    print("Generating model.")
    model = get_model(detectors[:, 0], detectors[:, 1], dl, dx, nx, x_0, nt, hash_fn=hash_fn)
    print("Model generation complete.")

    x = cp.linspace(-1, 1, nx, dtype=cp.float32)
    xs, ys = cp.meshgrid(x, x)

    p = 0.5 - (xs ** 2 + ys ** 2)
    p[p < 0] = 0

    plt.imshow(p.get())
    plt.show()

    timeseries = (model @ p.flatten()).get().reshape(-1, nt)

    plt.imshow(timeseries, aspect="auto")
    plt.show()

    y = timeseries[0:1, :]
    plt.plot(y.T)
    m1 = np.min(np.where(y != 0)[-1]) - 10
    m2 = np.max(np.where(y != 0)[-1]) + 10
    plt.xlim(m1, m2)
    plt.show()



#  Copyright (c) Thomas Else 2023.
#  License: MIT
import unittest

import patato as pat

import numpy as np

from patato import PAData


class TestModelBased(unittest.TestCase):
    def setUp(self) -> None:
        self.pa = PAData.from_hdf5("test_data.hdf5")

    def test_model_based_reconstruction(self):
        N, fov = (100, 100, 1), (0.025, 0.025, 1)
        mb = pat.ModelBasedReconstruction(N, fov, pa_example=self.pa, iter_lim=100)
        rec, _, _ = mb.run(self.pa.get_time_series(), self.pa)
        self.assertAlmostEqual(np.mean(rec.values), 17.165944017890236, 2)
        self.assertEqual(rec.shape, (1, 2, 1, 100, 100))



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import numpy as np
from scipy.sparse import csr_matrix, vstack

from os.path import join, exists, dirname
from .generate_model import calculate_element
from tqdm.auto import tqdm


def get_hash(*x):
    to_hash = []
    for y in x:
        if type(y) == np.ndarray:
            y = tuple(y.flatten())
        to_hash.append(y)
    h = hash(tuple(to_hash))
    return hex(np.uint64(h))


def generate_model(det_x, det_y, dl_0, dx, nx, x_0, nt):
    """

    Parameters
    ----------
    det_x
    det_y
    dl_0
    dl_1
    y_cutoff
    dx
    nx
    x_0
    nt

    Returns
    -------

    """
    # TODO: validate types.
    # TODO: allow non-square reconstruction areas.
    ntpixel = np.int32(4 * np.sqrt(2) * dx / dl_0)

    # Normalise the values. Convert to appropriate format.
    det_x = np.double(det_x)
    det_y = np.double(det_y)
    x_0 = np.double(x_0)
    dx = np.double(dx)

    matrices = []
    i = 0

    output = np.zeros((ntpixel * nx * nx), dtype=np.float64)
    indices = np.zeros((ntpixel * nx * nx), dtype=np.int32)

    positions = np.repeat(np.arange(nx * nx)[:, None], ntpixel, axis=-1).flatten()

    for detx, dety in tqdm(list(zip(det_x, det_y))):
        i += 1

        calculate_element(output, indices, nx, ntpixel, detx, dety, dl_0, x_0, dx)

        matrix = csr_matrix((output.flatten(), (indices.flatten(), positions)), shape=(nt, nx * nx))
        matrices.append(matrix)
    m = vstack(matrices)
    return m


def get_model(det_x, det_y, dl, dx, nx, x_0, nt,
              cache=True, hash_fn=None):
    det_x = det_x.astype(np.float64)
    det_y = det_y.astype(np.float64)
    dl = np.float64(dl)
    dx = np.float64(dx)
    nx = np.int32(nx)
    x_0 = np.float64(x_0)
    nt = np.int32(nt)
    print("Loading model")
    if hash_fn is None:
        hash_fn = get_hash
    if cache:
        h = hash_fn(det_x, det_y, dl, dx, nx, x_0, nt)
        model_folder = join(dirname(__file__), "models")
        filename = join(model_folder, h + ".npz")
        import scipy.sparse
        if exists(filename):
            mat = csr_matrix(scipy.sparse.load_npz(filename))
        else:
            mat = generate_model(det_x, det_y, dl, dx, nx, x_0, nt)
            scipy.sparse.save_npz(filename, mat.astype(np.float32).get(), compressed=False)
        return mat
    else:
        return generate_model(det_x, det_y, dl, dx, nx, x_0, nt)



#  Copyright (c) Thomas Else 2023.
#  License: MIT

import numpy as np


def load_ithera_irf(irf_file):
    irf_data = np.fromfile(irf_file, dtype=np.double)
    if irf_data[0] >= 1000:
        irf_data[0] -= 1000
    return irf_data
